/*
 *  YASS 1.11
 *  Copyright (C) 2004-2007
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>


/* 1) include utils macro */
#include "util.h"
/* 2) include global variables */
#include "global_var.h"
/* 3) the current file defs */
#include "display.h"
/* 4) other files */
#include "align.h"
#include "prdyn.h"
#include "tuple.h"
#include "proba.h"
#include "kword.h"

void Display_Alignements(MA * first_MA,
			int querysize,
			char *dataquery,
			char *dataqueryrev, char *queryname,
			char *datatext,
			int nbchunks, char **chunkname, int *chunksize,
			int *chunkstart)
{

  
  int nb_displayed_alignments = 0;
  MA *ma = NULL;

  /* [0] header if needed */

  if (gp_display == 2) {
    fprintf(OUTSTREAM,
	    "# Fields: Query id, Subject id, %% identity, alignment length, mismatches, gap openings, q. start, q. end, s. start, s. end, e-value, bit score\n");
    /* I hope this line is not copyrighted (because this is not mine)  ;-) */
  } else if (gp_display == 3) {
    fprintf(OUTSTREAM,
	    "# q. start,\tq. end,\ts. start,\ts. end,\tq. length,\ts.length,\tf/r\tbit score\te-value\n");
    /* Well this one is mine ... */
  } else if (gp_display == 4) {
    fprintf(OUTSTREAM,
	    "track name=\"%s\" description=\"%s\"\n",queryname,"empty");
    fprintf(OUTSTREAM,
	    "# s. id,\ts. start,\ts. end,\tstrand,\tq. start,\tq. end,\tbit score\te-value\t%%identity\n");
    
  }
  



    /* [1] Effective display of each Memorized Alignment */
    for (ma = first_MA;
	 ma != NULL && nb_displayed_alignments < gp_nbmaxlines;
	 ma = ma->next) {

	nb_displayed_alignments++;

#ifdef DEBUG
	DisplayMA(ma);
#endif

	switch (gp_display) {
	case 0:
	case 1:
	    /*  (1.1) alignment position repeat */
	    if (!(ma->reverse))
		fprintf(OUTSTREAM, "*(%d-%d)(%d-%d)"  ,
			BCOUNT + ma->left_pos_begin   ,
			BCOUNT + ma->left_pos_end  - 1,
			BCOUNT + ma->right_pos_begin  ,
			BCOUNT + ma->right_pos_end - 1);

	    else {
		fprintf(OUTSTREAM, "*(%d-%d)(%d-%d)",
			BCOUNT + querysize - ma->left_pos_begin - 1,
			BCOUNT + querysize - ma->left_pos_end      ,
			BCOUNT + ma->right_pos_begin               ,
			BCOUNT + ma->right_pos_end              - 1);
	    }
		/* (1.2) print details (size, evalue, etc ...) */
		fprintf(OUTSTREAM, " Ev: %g s: %d/%d %c\n",
			Evalue(gp_k_blast, gp_lambda_blast, gp_querysize,
			       gp_textsize, ma->blastscore),
			ma->left_pos_end - ma->left_pos_begin   ,
			ma->right_pos_end - ma->right_pos_begin ,
			(ma->reverse) ? 'r' : 'f');
	    
	    if (gp_display == 0)
		break;


	    /* (2.1) print chunk names */
	    fprintf(OUTSTREAM, "* %s / %s\n", queryname,chunkname[ma->i_chunk]);
	    /* (2.2) score bitscore and statistics */
	    fprintf(OUTSTREAM, "* score = %d : bitscore = %.2f\n",
		    (ma->blastscore), (float) BitScore(gp_k_blast,
						       gp_lambda_blast,
						       (ma->blastscore)));


	    
	    memset(gv_nb_mutations_per_codon1, '\0', 3 * sizeof(int));
	    memset(gv_nb_mutations_per_codon2, '\0', 3 * sizeof(int));
	    memset(gv_nb_transitions_transversions_indels, '\0', 3 * sizeof(int));
	    memset(gv_nb_triplet_count1, '\0', 64 * sizeof(int));
	    memset(gv_nb_triplet_count2, '\0', 64 * sizeof(int));
	    memset(gv_nb_non_mutated_triplet_count, '\0', 64 * sizeof(int));
	    memset(gv_nb_pair_of_triplets[0], '\0', 64 * 64 * sizeof(int));

	    alignment_SG_stats_on_MA((ma->reverse) ? dataqueryrev : dataquery,
				     datatext + chunkstart[ma->i_chunk], ma, 0,
				     gv_nb_transitions_transversions_indels, gv_nb_mutations_per_codon1, gv_nb_mutations_per_codon2);
	    
	    fprintf(OUTSTREAM,
		    /* "* mutations per triplet %d, %d, %d (%.2e) | ts : %d tv : %d | entropy : %g | mutual : %g\n\n", */
		    "* mutations per triplet %d, %d, %d (%.2e) | ts : %d tv : %d | entropy : %g\n\n",
		    gv_nb_mutations_per_codon1[0], gv_nb_mutations_per_codon1[1], gv_nb_mutations_per_codon1[2],
		    (float) (P_mutation_bias(gv_nb_mutations_per_codon1)), gv_nb_transitions_transversions_indels[0],
		    gv_nb_transitions_transversions_indels[1],
		    ma->entropy 
		    /* , ma->mutual */
		    );
		    
	    
	    /* (2.3) display alignment */
	    display_alignment_SG_on_MA((ma->reverse) ? dataqueryrev :
				       dataquery,
				       datatext + chunkstart[ma->i_chunk]
				       , ma);
	    
#ifdef DEBUG
	    DisplayMA(ma);
#endif
	    
	    break;		/* case 1: */

	case 2:
	    /* (3) blast like tabular format */
	    /* (3.1) fasta name */
	    fprintf(OUTSTREAM, "%s\t%s\t", queryname,
		    chunkname[ma->i_chunk]);

	    /* (3.2) stats analysis */

	    
	    memset(gv_nb_mutations_per_codon1, '\0', 3 * sizeof(int));
	    memset(gv_nb_mutations_per_codon2, '\0', 3 * sizeof(int));
	    memset(gv_nb_transitions_transversions_indels, '\0', 3 * sizeof(int));
	    memset(gv_nb_triplet_count1, '\0', 64 * sizeof(int));
	    memset(gv_nb_triplet_count2, '\0', 64 * sizeof(int));
	    memset(gv_nb_non_mutated_triplet_count, '\0', 64 * sizeof(int));
	    memset(gv_nb_pair_of_triplets[0], '\0', 64 * 64 * sizeof(int));

	    alignment_SG_stats_on_MA((ma->reverse) ? dataqueryrev : dataquery,
				     datatext + chunkstart[ma->i_chunk], ma, 0,
				     gv_nb_transitions_transversions_indels, gv_nb_mutations_per_codon1, gv_nb_mutations_per_codon2);

	    
	    /* (3.3) percent identity */
	    fprintf(OUTSTREAM, "%3.2f\t",
		    100 -
		    100.000 *
		    ((double)
		     (gv_nb_transitions_transversions_indels[0] + 
		      gv_nb_transitions_transversions_indels[1] +
		      gv_nb_transitions_transversions_indels[2]) / MAX((ma->left_pos_end -
					     ma->left_pos_begin ),
					    (ma->right_pos_end -
					     ma->right_pos_begin ))));
	    /* (3.4) size */
	    fprintf(OUTSTREAM, "%d\t",
		    MAX((ma->left_pos_end - ma->left_pos_begin),
			(ma->right_pos_end - ma->right_pos_begin)));
	    /* (3.5) mutations indels */
	    fprintf(OUTSTREAM, "%d\t%d\t", gv_nb_transitions_transversions_indels[0] + gv_nb_transitions_transversions_indels[1],
		    gv_nb_transitions_transversions_indels[2]);
	    /* (3.6) alignment position repeat */
	    if (!(ma->reverse))
		fprintf(OUTSTREAM, "%d\t%d\t%d\t%d\t",
			BCOUNT + ma->left_pos_begin,
			BCOUNT + ma->left_pos_end - 1,
			BCOUNT + ma->right_pos_begin,
			BCOUNT + ma->right_pos_end - 1);
	    else
		fprintf(OUTSTREAM, "%d\t%d\t%d\t%d\t", 
			BCOUNT + querysize - ma->left_pos_end      ,  /* inverted pair */
			BCOUNT + querysize - ma->left_pos_begin - 1,  
			BCOUNT + ma->right_pos_end - 1,	              /* inverted pair : blast database reversed */
			BCOUNT + ma->right_pos_begin );
	    fprintf(OUTSTREAM, "%.2g\t%.3g\n",
		    Evalue(gp_k_blast, gp_lambda_blast, gp_querysize,
			   gp_textsize, ma->blastscore),
		    BitScore(gp_k_blast, gp_lambda_blast,
			     ma->blastscore));
	    break;		/* case 2: */
	case 3:
	  /*  (4.1) alignment position repeat */
	  if (!(ma->reverse))
	    fprintf(OUTSTREAM, "%d\t%d\t%d\t%d",
		    BCOUNT + ma->left_pos_begin,
		    BCOUNT + ma->left_pos_end - 1,
		    BCOUNT + ma->right_pos_begin,
		    BCOUNT + ma->right_pos_end - 1);
	  else
	    fprintf(OUTSTREAM, "%d\t%d\t%d\t%d",
		    BCOUNT + querysize - ma->left_pos_end,       /* inverted pair */
		    BCOUNT + querysize - ma->left_pos_begin - 1,
		    BCOUNT + ma->right_pos_begin,                /* normal pair */
		    BCOUNT + ma->right_pos_end              - 1);

	    /* (4.2) print details (size, f/b , bitscore , evalue ) */
	  fprintf(OUTSTREAM, "\t%d\t%d\t%c\t%2g\t%3g\n",
		  ma->left_pos_end  - ma->left_pos_begin,
		  ma->right_pos_end - ma->right_pos_begin,
		  (ma->reverse) ? 'r' : 'f',
		  BitScore(gp_k_blast, gp_lambda_blast, ma->blastscore),
		  Evalue(gp_k_blast, gp_lambda_blast, gp_querysize, gp_textsize, ma->blastscore)
		  );
	  break;

	case 4:
	  /*  (5.0) bed file name */
	  fprintf(OUTSTREAM, "%s\t", queryname /* chunkname[ma->i_chunk] */);
	  
	  /*  (5.1) alignment position repeat */
	  if (!(ma->reverse))
	    fprintf(OUTSTREAM, "%d\t%d\t%c\t%d\t%d\t",
		    BCOUNT + ma->left_pos_begin,
		    BCOUNT + ma->left_pos_end - 1,
		    '+',
		    BCOUNT + ma->right_pos_begin,
		    BCOUNT + ma->right_pos_end - 1);
	  else
	    fprintf(OUTSTREAM, "%d\t%d\t%c\t%d\t%d\t",
		    BCOUNT + querysize - ma->left_pos_end,       /* inverted pair */
		    BCOUNT + querysize - ma->left_pos_begin - 1,
		    '-',
		    BCOUNT + ma->right_pos_begin,                /* normal pair */
		    BCOUNT + ma->right_pos_end              - 1);

	  /* (5.2) bitscore / evalue */
	  fprintf(OUTSTREAM, "%.3g\t%.2g\t",
		  BitScore(gp_k_blast, gp_lambda_blast,
			   ma->blastscore),
		  Evalue(gp_k_blast, gp_lambda_blast, gp_querysize,
			 gp_textsize, ma->blastscore)
		  );
	  


	  /* (5.3) stats analysis */
	  memset(gv_nb_mutations_per_codon1, '\0', 3 * sizeof(int));
	  memset(gv_nb_mutations_per_codon2, '\0', 3 * sizeof(int));
	  memset(gv_nb_transitions_transversions_indels, '\0', 3 * sizeof(int));
	  memset(gv_nb_triplet_count1, '\0', 64 * sizeof(int));
	  memset(gv_nb_triplet_count2, '\0', 64 * sizeof(int));
	  memset(gv_nb_non_mutated_triplet_count, '\0', 64 * sizeof(int));
	  memset(gv_nb_pair_of_triplets[0], '\0', 64 * 64 * sizeof(int));
	  
	  alignment_SG_stats_on_MA((ma->reverse) ? dataqueryrev : dataquery,
				   datatext + chunkstart[ma->i_chunk], ma, 0,
				   gv_nb_transitions_transversions_indels, gv_nb_mutations_per_codon1, gv_nb_mutations_per_codon2);
	  
	  
	  /* (5.4) percent identity */
	  fprintf(OUTSTREAM, "%3.2f\n",
		  100 -
		  100.000 *
		  ((double)
		   (gv_nb_transitions_transversions_indels[0] + 
		    gv_nb_transitions_transversions_indels[1] +
		    gv_nb_transitions_transversions_indels[2]) / MAX((ma->left_pos_end -
								      ma->left_pos_begin ),
								     (ma->right_pos_end -
								      ma->right_pos_begin ))));
	  break;
	} /* switch */
    } /* for */
}




void Display_Stats(Feature * feature /* feature[2] */)
{
    fprintf(stderr, "* Statistical parameters\n");
    fprintf(stderr, "  lambda = %g\n", gp_lambda_blast);
    fprintf(stderr, "  K      = %g\n", gp_k_blast);
    fprintf(stderr, "  rho    = %d\n", gp_rho_stat);
    fprintf(stderr, "  delta  = %d\n", gp_delta_stat);

#ifdef STATS
    fprintf(stderr, "  ----\n");
    /* preprocess */
    fprintf(stderr, "  # keys removed     =   f:%9d   r:%9d   total:%9d \n", 
	    feature[0].nb_keys_removed,
	    feature[1].nb_keys_removed,
	    feature[0].nb_keys_removed + feature[1].nb_keys_removed);
    /* hits */
    fprintf(stderr, "  -\n");
    fprintf(stderr, "  # single hit       =   f:%9d   r:%9d   total:%9d \n", 
	    feature[0].nb_seeds, 
	    feature[1].nb_seeds, 
	    feature[0].nb_seeds + feature[1].nb_seeds);
    fprintf(stderr, "  # chains built     =   f:%9d   r:%9d   total:%9d \n", 
	    feature[0].nb_chains_built , 
	    feature[1].nb_chains_built ,
	    feature[0].nb_chains_built + feature[1].nb_chains_built
	    );
    fprintf(stderr, "  # align. tested    =   f:%9d   r:%9d   total:%9d \n", 
	    feature[0].nb_chains_tested ,
	    feature[1].nb_chains_tested ,
	    feature[0].nb_chains_tested + feature[1].nb_chains_tested
	    );
    fprintf(stderr, "  # align. detected  =   f:%9d   r:%9d   total:%9d \n", 
	    feature[0].nb_ma , 
	    feature[1].nb_ma , 
	    feature[0].nb_ma + feature[1].nb_ma
	    );
    /* regroup */
    fprintf(stderr, "  -\n");
    fprintf(stderr, "  # grouping tests   =   f:%9d   r:%9d   total:%9d \n",
	    feature[0].nb_postprocessed_grouping_tests,
	    feature[1].nb_postprocessed_grouping_tests,
	    feature[0].nb_postprocessed_grouping_tests + feature[1].nb_postprocessed_grouping_tests
	    );
    fprintf(stderr, "  # align. linked    =   f:%9d   r:%9d   total:%9d \n",
	    feature[0].nb_postprocessed_grouping_links,
	    feature[1].nb_postprocessed_grouping_links,
	    feature[0].nb_postprocessed_grouping_links + feature[1].nb_postprocessed_grouping_links
	    );
    fprintf(stderr, "  # align. postpros  =   f:%9d   r:%9d   total:%9d \n",
	    feature[0].nb_postprocessed_ma,
	    feature[1].nb_postprocessed_ma,
	    feature[0].nb_postprocessed_ma + feature[1].nb_postprocessed_ma
	    );

    /* cpu-times */
    fprintf(stderr, "  ----\n");
    fprintf(stderr, "  pre-proc. cpu-time =   f:%8.2fs   r:%8.2fs   total:%8.2fs [%d%%]\n",
	    feature[0].clock_pre  /CLOCKS_PER_SEC,
	    feature[1].clock_pre  /CLOCKS_PER_SEC,
	    (feature[0].clock_pre   + feature[1].clock_pre  )/CLOCKS_PER_SEC,
	    (int)(
	    (feature[0].clock_pre   + feature[1].clock_pre  )/
	    (feature[0].clock_pre   + feature[1].clock_pre   +
	     feature[0].clock_chain + feature[1].clock_chain +
	     feature[0].clock_align + feature[1].clock_align +
	     feature[0].clock_post  + feature[1].clock_post  + 1e-6)*100)
	    );
    fprintf(stderr, "  chain.    cpu-time =   f:%8.2fs   r:%8.2fs   total:%8.2fs [%d%%]\n",
	    feature[0].clock_chain/CLOCKS_PER_SEC,
	    feature[1].clock_chain/CLOCKS_PER_SEC,
	    (feature[0].clock_chain + feature[1].clock_chain)/CLOCKS_PER_SEC,
	    (int)(
	    (feature[0].clock_chain + feature[1].clock_chain)/
	    (feature[0].clock_pre   + feature[1].clock_pre   +
	     feature[0].clock_chain + feature[1].clock_chain +
	     feature[0].clock_align + feature[1].clock_align +
	     feature[0].clock_post  + feature[1].clock_post  + 1e-6)*100)
	    );
    fprintf(stderr, "  align.    cpu-time =   f:%8.2fs   r:%8.2fs   total:%8.2fs [%d%%]\n",
	    feature[0].clock_align/CLOCKS_PER_SEC,
	    feature[1].clock_align/CLOCKS_PER_SEC,
	    (feature[0].clock_align + feature[1].clock_align)/CLOCKS_PER_SEC,
	     (int)(
	    (feature[0].clock_align + feature[1].clock_align)/
	    (feature[0].clock_pre   + feature[1].clock_pre   +
	     feature[0].clock_chain + feature[1].clock_chain +
	     feature[0].clock_align + feature[1].clock_align +
	     feature[0].clock_post  + feature[1].clock_post   + 1e-6)*100)
	    );
    fprintf(stderr, "  post-proc.cpu-time =   f:%8.2fs   r:%8.2fs   total:%8.2fs [%d%%]\n",
	    feature[0].clock_post/CLOCKS_PER_SEC,
	    feature[1].clock_post/CLOCKS_PER_SEC,
	    (feature[0].clock_post + feature[1].clock_post)/CLOCKS_PER_SEC,
	     (int)(
	    (feature[0].clock_post  + feature[1].clock_post )/
	    (feature[0].clock_pre   + feature[1].clock_pre   +
	     feature[0].clock_chain + feature[1].clock_chain +
	     feature[0].clock_align + feature[1].clock_align +
	     feature[0].clock_post  + feature[1].clock_post   + 1e-6)*100)
	    );    

#ifdef DEBUG
    DisplayHistoScore(feature);
#endif

 /*
    
    if (clock_align + clock_chain + clock_post == 0)
      clock_align = clock_chain = clock_post = 1;

    fprintf(stderr, "       chaining  time = %d s [ %d%%]\n",
	    (int) (0.5 +
		   gv_time_spent * (clock_chain /
				 (clock_align + clock_chain + clock_post)))
	    ,
	    (int) (0.5 +
		   100 * (clock_chain /
			  (clock_align + clock_chain + clock_post)))
	);
    fprintf(stderr, "       alignment time = %d s [ %d%%]\n",
	    (int) (0.5 +
		   gv_time_spent * (clock_align /
				 (clock_align + clock_chain + clock_post)))
	    ,
	    (int) (0.5 +
		   100 * (clock_align /
			  (clock_align + clock_chain + clock_post)))
	);
    fprintf(stderr, " post-processing time = %d s [ %d%%]\n",
	    (int) (0.5 +
		   gv_time_spent * (clock_post /
				 (clock_align + clock_chain + clock_post)))
	    ,
	    (int) (0.5 +
		   100 * (clock_post /
			  (clock_align + clock_chain + clock_post)))
	);
    */

    fprintf(stderr, "     real-time spent = %ds\n", gv_time_spent);

#endif

}




void Display_Params()
{
    int i;
    fprintf(stderr, "\n");
    fprintf(stderr, RULE);

    /* filenames */
    if (gp_nbfiles <= 1) {
	fprintf(stderr, "  DNA file : ");
	fprintf(stderr, "\"%s\", %d chunk(s), selected chunk: %d (size : %d bp, GC: %2.2f%%)\n", gp_files[0], gp_nbchunks_query, gp_selection_fasta, gp_querysize_chunk,  100*(gp_freq_letters[0][1/*C*/] +gp_freq_letters[0][2/*G*/]));
    } else {
	fprintf(stderr, "  first DNA file : ");
	fprintf(stderr, "\"%s\", %d chunk(s), selected chunk: %d (size : %d bp, GC: %2.2f%%)\n", gp_files[0], gp_nbchunks_query, gp_selection_fasta, gp_querysize_chunk,  100*(gp_freq_letters[0][1/*C*/] +gp_freq_letters[0][2/*G*/]));
	fprintf(stderr, "  second DNA file : ");
	fprintf(stderr, "\"%s\"  %d chunk(s), (overall size : %d bp,  GC: %2.2f%%)\n\n", gp_files[1], gp_nbchunks_text, gp_textsize, 100*(gp_freq_letters[1][1/*C*/] +gp_freq_letters[1][2/*G*/]));
    }

    /* parameters */
    fprintf(stderr, "* Parameters\n");
    for (i = 0; i < gp_nb_seeds; i++)
	fprintf(stderr,
		"  seed[%d]: pattern = \"%s\", weight = %d bits, span = %d bases\n",
		i, gp_motifs[i], gp_seeds_bitweight[i], gp_seeds_span[i]
		);
    fprintf(stderr,
	    "  cost (match,mismatch,transition,other) = %d,%d,%d,%d\n",
	    gp_costs[0], gp_costs[1], gp_costs[2], gp_costs[3]
	    );
    fprintf(stderr, 
	    "  cost gaps (opening,continous) = %d,%d\n",
	    gp_cost_gap_opened, gp_cost_gap_continued
	    );
    fprintf(stderr,
	    "  x-drop cost = %d\n",
	    gp_xdrop
	    );
    fprintf(stderr, 
	    "  mutation rate = %d%%\n", 
	    gp_mutations_percent
	    );
    fprintf(stderr, 
	    "  indels rate = %d%%\n",
	    gp_indels_percent
	    );
    fprintf(stderr, 
	    "  tolerance  = %d%%\n", 
	    gp_alpha_percent
	    );
    fprintf(stderr, 
	    "  minimum score seeds = %d\n",
	    gp_minimum_score_seeds
	    );
    fprintf(stderr, 
	    "  minimal score = %d\n",
	    gp_minscore
	    );
}



/* Dot progress bar : 
 * "launch and lets take a coffee"(TM)  progress bar
 */


void Display_Progress(unsigned int pos, unsigned int maxpos,Feature  *feature)
{
    if (gp_reverse == 2) {
	    if (feature->last_point * maxpos < pos * (NBL / 2)) {
		while (feature->last_point * maxpos < pos * (NBL / 2)
		       && feature->last_point < NBL) {
		    fprintf(stderr, ".");
		    fflush(NULL);
		    (feature->last_point)++;
		}
	    }
    } else {
	if (feature->last_point * maxpos < pos * (NBL)) {
	    while (feature->last_point * maxpos < pos * (NBL) && feature->last_point < NBL) {
		fprintf(stderr, ".");
		fflush(NULL);
		(feature->last_point)++;
	    }
	}
    }
}


void DisplayHistoScore(Feature * feature /* feature[2] */) {
  int i;
  
  fprintf(stderr," Scoring Histogram : \n");
  for ( i = 0 ; i < sizeof(int)*8-2 ; i++ ) {
   
   
    
    int lines_f = 20;
    int lines_r = 20;
    
    fprintf(stderr,"\t score 2^%2d to 2^%2d \t",i,i+1);
    
    if (feature[0].MAcount){
      int count_f = feature[0].MAcount[i];
      while(count_f > 0){
	count_f >>= 1;
	lines_f--;
	fprintf(stderr,".");
      }
      while(lines_f > 0){
	lines_f--;
	fprintf(stderr," ");
      }
      fprintf(stderr,"(%8d)",feature[0].MAcount[i]);
    }else {
      fprintf(stderr,
	      "                                  "
	      );
    }
    
    if (feature[1].MAcount){
      int count_r = feature[1].MAcount[i];
      while(count_r > 0){
	count_r >>= 1;
	lines_r--;
	fprintf(stderr,".");
      }
      while(lines_r > 0){
	lines_r--;
	fprintf(stderr," ");
      }
      fprintf(stderr,"(%8d)\n",feature[1].MAcount[i]);
    }else {
      fprintf(stderr,
	      "                                  "
	      );
    }
  }
}



void DisplaySubstitutionMatrix() {
  int i,j;
  for(i=0;i<4;i++) {
    for(j=0;j<4;j++) {
      fprintf(stderr,"%2d",gp_substitution_matrix[i][j]);
      if (j==3)
	fprintf(stderr,"\n");
      else 
	fprintf(stderr,",");
    }
  }
}
