/*
 *  YASS 1.11
 *  Copyright (C) 2004-2007
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>

/* 1) include utils macro */
#include "util.h"
/* 2) include global variables */
/* 3) the current file defs */
#include "global_var.h"
/* 4) other files */
#include "tuple.h"

/********************
 * GLOBALVARIABLES  *
 ********************/



int SUBMATRIXTABLE[4][4] = {	
    /* Match, Transversion, Transition, other */
    {1, -3, -2, -3},/* Old blast like */
    {3, -3, -2, -3},/* */
    {3, -4, -3, -4},/* Human/Mouse like */
    {5, -4, -3, -4} /* EDNAFULL like */
};

int INDELSTABLE[4][2] = {	/* +1 -3 */
    {-8, -2},
    /* +3 -2 */
    {-10, -4},
    /* +5 -4 */
    {-16, -4},
    /* +4 -3 */
    {-16, -6}
};



/* [1]  MODIFIED WHEN STARTING PROGRAM 
 *========================================*/


/* 1.1 parametres et flags modifis par l'utilisateur */

int    gp_cost_gap_opened     = -16;
int    gp_cost_gap_continued  = -4;
int    gp_costs[4]            = {5,-4,-2,-4}; /* match/mismatch/transition/other*/
int ** gp_substitution_matrix = NULL;
int    gp_matrix = 3;
int    gp_cost_max_substitution_matrix = 0; /* scoring matrix maximal absolute value */


double gp_k_blast = -1.0;
double gp_lambda_blast = -1.0;
int    gp_xdrop = 25;           

double gp_t = 1e-3;		/* t :  max repeats ratio */
char *gp_motifs[MAX_SEED];
int gp_nb_seeds;
int gp_seeds_span_max;
int gp_seeds_span_min;
int gp_seeds_span[MAX_SEED];
int gp_seeds_bitweight[MAX_SEED];
int gp_seeds_bitmask[MAX_SEED];  /* binary seed mask */ 
int *gp_automata = NULL;	/* increment overlap */

int gp_minimum_score_seeds = 12;
int gp_minscore = 1;

int    gp_mutations_percent = 25;
double gp_mutations = 0.25;
int    gp_indels_percent = 8;
double gp_indels = 0.08;
int    gp_alpha_percent = 5;
double gp_alpha = 0.05;

int gp_reverse = 2;
int gp_display = 0;
int gp_nbmaxlines = 100000; /* maximal number of alignements in the output */
int gp_distdiag = 16;
int gp_lowercase = 0;
double gp_entropy_min = 2.80;
double gp_expectation_value = 10.00;

/* post-processing */
int     gp_win_min = 16;
int     gp_win_max = 64 * KILO;
double  gp_win_mul = 4.0;

/* sort function index */
int        gp_sortcriterion      = 0;
SortCrit * gp_sortcriterion_func = SortCriterionScore;


/* statistical variables */
int gp_rho_stat = 0;
int gp_delta_stat = 0;

/* files */
int  gp_nbfiles = 0;
char gp_files[2][256];
int  gp_selection_fasta = 1;

/* data tables */
char * gp_query       = NULL, *gp_text           = NULL;
char * gp_query_chunk = NULL, *gp_queryrev_chunk = NULL;
int    gp_querysize   = 0,     gp_textsize       = 0,    gp_querysize_chunk = 0;

/* first file chunks */
int    gp_nbchunks_query  = 0;
char **gp_chunkname_query = NULL;
int   *gp_chunksize_query = NULL;
int   *gp_chunkstrt_query = NULL;

/* second file chunks */
int     gp_nbchunks_text = 0;
char **gp_chunkname_text = NULL;
int   *gp_chunksize_text = NULL;
int   *gp_chunkstrt_text = NULL;
       

/* statistical variables */
int gp_nb_letters[2][4] =   { { 0, 0, 0, 0
                            },{ 0, 0, 0, 0
			    } }; /* number of 'A','T','G','C' for each file */

double ** gp_freq_letters /* [2][4] */ = NULL ;
                           /* frequency of 'A','T','G','C' for each file */

int gp_nb_triplets[2][64] = { { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
                            },{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
			    } }; /* number of 'AAA','AAT', ... 'CCC' for each file */


double ** gp_freq_background;        /* 4x4 */
double ** gp_freq_tripletbackground; /* 64x64 */

char lookup[32] =
    { 'A', 'C', 'G', 'T',
      'R', 'Y', 'N', 'U', 
      'K', 'W', 'S', 'M',
      'V', 'H', 'D', 'B',
      'a', 'c', 'g', 't',
      'r', 'y', 'n', 'u', 
      'k', 'w', 's', 'm',
      'v', 'h', 'd', 'b' 
    };


int backcode[32] = {
  0, 1, 2, 3,
  0, 1, 0, 1,
  0, 0, 0, 0,
  0, 0, 0, 0,
  0, 1, 2, 3,
  0, 1, 0, 1,
  0, 0, 0, 0,
  0, 0, 0, 0
};

char complement[32] = {
   3, 2, 1, 0,
   5, 4, 7, 0,
   9, 8,11,10,
  15,14,13,12,
  19,18,17,16,
  21,20,23,16,
  25,24,27,26,
  29,28,31,30
};

int unindexable[32] = {
  0, 0, 0, 0,
  0, 0, 1, 0,
  1, 1, 1, 1,
  1, 1, 1, 1,
  0, 0, 0, 0,
  0, 0, 1, 0,
  1, 1, 1, 1,
  1, 1, 1, 1
};

SortCrit* sortcriteria[NBSORTCRITERIA] = {
  SortCriterionScore,
  SortCriterionEntropy,
  SortCriterionMutual,
  SortCriterionScoreWithEntropy,
  SortCriterionQueryBegin
};




#ifdef LOW_MEMORY
int gp_hash_size    = 65536;
int gp_hash_element = 16384;
#endif 

#ifdef MEM_ALLOCATED
unsigned long int gv_mem_allocated     = 0;
unsigned long int gp_max_mem_allocated = 1024*MEGA;
#endif


/* [2]   MODIFIED WHEN RUNNING ...
 *================================*/

FILE * gv_outstream = NULL;

int    gv_thread_result = 0;

int  * gv_delta_shift = NULL;

int gv_nb_mutations_per_codon1[3];
int gv_nb_mutations_per_codon2[3];
int gv_nb_transitions_transversions_indels[3];
int gv_nb_triplet_count1[64];
int gv_nb_triplet_count2[64];
int gv_nb_non_mutated_triplet_count[64];
int ** gv_nb_pair_of_triplets = NULL;



/* statistics on time spent */
#ifdef STATS
int gv_time_spent = 0;
#endif
