/*
 *  YASS 1.11
 *  Copyright (C) 2004-2007
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef  __GLOBAL_VAR_H_
#define  __GLOBAL_VAR_H_
#include "util.h"
#include "tuple.h"




/********************
 * GLOBALVARIABLES  *
 ********************/


/* [1]  MODIFIED WHEN STARTING PROGRAM 
 *========================================*/

/* parameters and flags */
extern char *gp_motifs[MAX_SEED];
extern int gp_seeds_span[MAX_SEED];
extern int gp_seeds_bitweight[MAX_SEED];
extern int gp_seeds_bitmask[MAX_SEED];  /* binary seed mask */ 
extern int gp_seeds_span_min;
extern int gp_seeds_span_max;
extern int gp_nb_seeds;
extern int *gp_automata;

extern int gp_minimum_score_seeds;
extern int gp_minscore;
extern int gp_mutations_percent;
extern double gp_mutations;
extern int gp_indels_percent;
extern double gp_indels;
extern int gp_alpha_percent;
extern double gp_alpha;

extern double gp_entropy_min;
extern double gp_expectation_value;

extern int gp_reverse;
extern int gp_display;
extern int gp_nbmaxlines;
extern int gp_distdiag;
extern int gp_lowercase;
extern double gp_t;

/* statisticals parameters */
extern int gp_rho_stat;
extern int gp_delta_stat;


/* post-processing */
extern int gp_win_min;
extern int gp_win_max;
extern double gp_win_mul;

/* sort function index */
extern int        gp_sortcriterion;
extern SortCrit * gp_sortcriterion_func;

/* files number et filenames */
extern int gp_nbfiles;
extern char gp_files[2][256];
extern int gp_selection_fasta;


/* cost */
extern int SUBMATRIXTABLE[4][4];
extern int INDELSTABLE[4][2];
extern int gp_cost_gap_opened;
extern int gp_cost_gap_continued;
extern int gp_costs[4];
extern int ** gp_substitution_matrix;
extern int gp_cost_max_substitution_matrix;
extern int gp_matrix;
extern double gp_k_blast;
extern double gp_lambda_blast;
extern int gp_xdrop;

/* data tables and their sizes  */
extern char *gp_query;
extern int gp_querysize;

extern char *gp_query_chunk;
extern char *gp_queryrev_chunk;
extern int gp_querysize_chunk;

extern char *gp_text;
extern int gp_textsize;

/* first file chunks */
extern int     gp_nbchunks_query;
extern char ** gp_chunkname_query;
extern int   * gp_chunksize_query;
extern int   * gp_chunkstrt_query;

/* second file chunks */
extern int     gp_nbchunks_text;
extern char ** gp_chunkname_text;
extern int   * gp_chunksize_text;
extern int   * gp_chunkstrt_text;
      


/* statistical variables */
extern int gp_nb_letters[2][4];
extern double ** gp_freq_letters /*[2][4]*/;
extern int gp_nb_triplets[2][64];
extern double ** gp_freq_background;        /* 4x4 */
extern double ** gp_freq_tripletbackground; /* 64x64 */


/* lockup/code reverals and other stuff */
extern char lookup[32];
extern int  backcode[32];
extern char complement[32];
extern int  unindexable[32];
extern SortCrit* sortcriteria[NBSORTCRITERIA] ;

/* size of hash table */
extern int gp_hash_size;
extern int gp_hash_element;

#ifdef MEM_ALLOCATED
extern unsigned long gv_mem_allocated;
extern unsigned long gp_max_mem_allocated;
#endif

 
/* [2]   MODIFIED WHEN RUNNING ...
 *================================*/
extern FILE *gv_outstream;

extern int   gv_thread_result;

extern int  *gv_delta_shift;

extern int gv_nb_mutations_per_codon1[3];
extern int gv_nb_mutations_per_codon2[3];
extern int gv_nb_transitions_transversions_indels[3];
extern int gv_nb_triplet_count1[64];
extern int gv_nb_triplet_count2[64];
extern int gv_nb_non_mutated_triplet_count[64];

extern int ** gv_nb_pair_of_triplets;

/* statistics on time spent */
#ifdef STATS
extern int gv_time_spent;
#endif

#endif
