/*
 *  YASS 1.11
 *  Copyright (C) 2004-2007
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef __HASH_H__
#define __HASH_H__

/*
 *  Structure interne a la biliotheque
 * 
 *
 */
typedef struct _Hash_Element HashElement;
struct _Hash_Element {
  int  diagonal;
  int  value;

  HashElement *next;        /*Suivant dans la table de hach (evite les collitions*/
  HashElement *prev;

  HashElement *young_next;  /*Suivant dans la liste des insertions */
  HashElement *old_next;

};


typedef struct _Table_hash {
 
  int size;
  int mask;
  HashElement **element;
  
  HashElement *firstElement;  /*Premier element alloue (Pour liberer la memoire)*/
  HashElement *elementFree;   /*elements libres */
  HashElement *element_old;   /*premier element inserer dans la table de hash*/
  HashElement *element_young; /*dernier   element inserer dans la table */
  
}Table_hash;


/*
 *
 *  Initialisation de la table de Table
 *  de taille SIZE
 *
 *  Plus la taille est grande plus l'algorithme est
 *  rapide mais prends beaucoup de memoire.
 *
 */
Table_hash initHash(int size,int nbreElement);


/*
 *   Recupere la valeur d une clef
 *   Retourne:  -1 si le nombre n'existe pas (voir existHashKey)
 *              la valeur
 *
 */
int getHashValue (Table_hash *table,int key);

int getHashValueMin (Table_hash *table,int diagonal,int valueMin);


/*
 *  Etablie l'existance de la clef
 *  Return: -1 si la clef n'existe pas
 *          1 si elle existe
 *
 */
int existHashDiag (Table_hash *table,int key);


/*
 * Insere rapidement une clef et sa valeur
 * Attention: Il peut y avoir des toublons
 *
 */
int putHashValue (Table_hash *table,int key, int value);


/*
 * Insere une clef unique et sa valeur
 * Si jamais la clef existe la valeur sera ecrase
 *
 */
int putHashOneDiagValue (Table_hash *table,int key, int value) ;

int delHash (Table_hash *table,int diagonal);

/*
 *
 * Affiche toutes la table de hack
 *
 */
int hashDisplay (Table_hash *table);


/*
 * 
 * supprime la table de hash
 *
 */
int freeHash (Table_hash *table);

/*
 * 
 * reset: enleve tous les elements de la table pour les mettre dans free
 *
 */
int resetHash (Table_hash *table);

#endif
