/*
 *  YASS 1.11
 *  Copyright (C) 2004-2007
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>

/* 1) include utils macro */
#include "util.h"
/* 2) include global variables */
#include "global_var.h"
/* 3) the current file defs */
#include "kword.h"




/*
 *   Build a  DNA bases table  (code 0..3) from a DNA file
 * 
 * - Input  : "filename"                                    : DNA file name
 * - Output : "p_data"                                      : return the allocated data table pointer 
 *            "p_datasize"                                  : return the allocated data table size 
 *            "p_nbchunks","p_chunkname", and "p_chunksize" : return respectively the number of chunks fond in the multifasta file, their names and their size
 *                                                            (set to NULL if you do not need these informations
 */



#define WORDCOUNTANDSTAT(count,letter,word) {(word) <<= 2;  (word) |= (letter); (word) &= 0xfc00003f;  if (((word) & 0x3f) == (word)) count[(word)]++;}


int CreateData(/* in */  char *filename,
	       /* out */ char **p_data,     /* out */ int *p_datasize, 
	       /* out */ int *p_nbchunks,   /* out */ char ***p_chunkname, 
	       /* out */ int **p_chunksize, /* out */ int **p_chunkstart,
	       /* out */ int * nb_letters  /* [4]  */ , 
	       /* out */ int * nb_triplets /* [64] */ )
{
  
  int i = 0, j = 0;
  int car = 0;
  int datasize = 0;
  int cumulength = 0;
  char *data = NULL;
  FILE *file = NULL;
  
  int nbchunks = 0;
  char **chunkname = NULL;
  int *chunksize = NULL;
  int *chunkstart = NULL;
  unsigned int word = 0xfc000000;
  
  int line = 1,column = 1;;
  /* (1) count number of DNA bases availaible */


/*
 *      A --> adenosine           K --> G T (keto)
 *      T --> thymidine           M --> A C (amino)
 *      G --> guanine             S --> G C (strong)
 *      C --> cytidine            W --> A T (weak)
 *      R --> G A (purine)	  V --> G C A
 *      Y --> T C (pyrimidine)	  B --> G T C
 *	N --> A G C T (any)       D --> G A T
 *      U --> uridine	  	  H --> A C T
 */


    if (!(file = fopen(filename, "r"))) {
      fprintf(stderr,"* Error : unable to open \"%s\": errno = %i \n", filename,errno);
      exit(0);
    }

    /* comment */
    while ((car = fgetc(file)) != EOF) {
      
      if (car > 127){
	fprintf(stderr,"* Error : file \"%s\" : non ASCII character \"0x%2X\" at line %d, column %d\n",filename,(unsigned)car,line,column);
      }
      
      switch (car) {
      case 'A':
      case 'a':
      case 'C':
      case 'c':
      case 'G':
      case 'g':
      case 'T':
      case 't':
      case 'R':
      case 'r':
      case 'Y':
      case 'y':
      case 'N':
      case 'n':
      case 'U':
      case 'u':
      case 'K':
      case 'k':
      case 'M':
      case 'm':
      case 'S':
      case 's':
      case 'W':
      case 'w':
      case 'V':
      case 'v':
      case 'B':
      case 'b':
      case 'D':
      case 'd':
      case 'H':
      case 'h':
	column++;
	datasize++;
	break;	
      case '>':
      case '#':
      case ':':
	column++;
	nbchunks++;
	do {
	  /* seek eol */
	  car = fgetc(file);
	  column++;
	} while (car >= 0x20 && car != EOF && car != '\n' && car < 128);
	break;
      case '\n':
	column = 1;
	line++;
	break;
	
      default:
	if (car >= 0x20 && car != EOF && car != '\n' && car < 128)
	  column++;
	else {
	  column = 1;
	  line++;
	}
	break;
      }	/*switch */
    } /* while */
    

    if (!datasize) {
      fprintf(stderr,"* Error : empty file \"%s\" ",filename);
      exit(0);
    }

    fclose(file);

    /* (2) allocate one memory blok for the DNA data */

    data = (char *) MALLOC((unsigned) datasize*sizeof(char));
    ASSERT(data, CreateData);

    if (nbchunks > 0) {
	chunkname = (char **) MALLOC((unsigned) nbchunks* sizeof(char *));
	ASSERT(chunkname, CreateData);
	memset(chunkname,0,(unsigned) nbchunks* sizeof(char *));
	chunksize = (int *) MALLOC((unsigned) nbchunks* sizeof(int));
	ASSERT(chunksize, Createdata);
	memset(chunksize,0,(unsigned) nbchunks* sizeof(int));
	chunkstart = (int *) MALLOC((unsigned) nbchunks* sizeof(int));
	ASSERT(chunkstart, Createdata);
	memset(chunkstart,0,(unsigned) nbchunks* sizeof(int));
    }


    /* (3) fill the DNA data */
    if (!(file = fopen(filename, "r")))
	return -3;
    /* description of sentance */
    while ((car = fgetc(file)) != EOF) {
	if (car == EOF)
	    break;
	switch (car) {

	case 'A':
	    data[i++] = (char) 0;  nb_letters[0]++;
	    WORDCOUNTANDSTAT(nb_triplets,0,word);
	    break;

	case 'C':
	    data[i++] = (char) 1;  nb_letters[1]++;
	    WORDCOUNTANDSTAT(nb_triplets,1,word);
	    break;

	case 'G':
	    data[i++] = (char) 2;  nb_letters[2]++;
	    WORDCOUNTANDSTAT(nb_triplets,2,word);
	    break;

	case 'T':
	    data[i++] = (char) 3;  nb_letters[3]++;
	    WORDCOUNTANDSTAT(nb_triplets,3,word);
	    break;

	case 'R':
	    data[i++] = (char) 4;
	    break;

	case 'Y':
	    data[i++] = (char) 5;
	    break;

	case 'N':
	    data[i++] = (char) 6;
	    break;

	case 'U':
	    data[i++] = (char) 7;
	    WORDCOUNTANDSTAT(nb_triplets,3,word);
	    break;

	case 'K':
	    data[i++] = (char) 8;
	    break;

	case 'W':
	    data[i++] = (char) 9;
	    break;

	case 'S':
	    data[i++] = (char) 10;
	    break;

	case 'M':
	    data[i++] = (char) 11;
	    break;

	case 'V':
	    data[i++] = (char) 12;
	    break;

	case 'H':
	    data[i++] = (char) 13;
	    break;

	case 'D':
	    data[i++] = (char) 14;
	    break;

	case 'B':
	    data[i++] = (char) 15;
	    break;

	case 'a':
	    data[i++] = (char) 16;  nb_letters[0]++;
	    WORDCOUNTANDSTAT(nb_triplets,0,word);
	    break;

	case 'c':
	    data[i++] = (char) 17;  nb_letters[1]++;
	    WORDCOUNTANDSTAT(nb_triplets,1,word);
	    break;

	case 'g':
	    data[i++] = (char) 18;  nb_letters[2]++;
	    WORDCOUNTANDSTAT(nb_triplets,2,word);
	    break;

	case 't':
	    data[i++] = (char) 19;  nb_letters[3]++;
	    WORDCOUNTANDSTAT(nb_triplets,3,word);
	    break;

	case 'r':
	    data[i++] = (char) 20;
	    break;

	case 'y':
	    data[i++] = (char) 21;
	    break;

	case 'n':
	    data[i++] = (char) 22;
	    break;

	case 'u':
	    data[i++] = (char) 23;
	    WORDCOUNTANDSTAT(nb_triplets,3,word);
	    break;

	case 'k':
	    data[i++] = (char) 24;
	    break;

	case 'w':
	    data[i++] = (char) 25;
	    break;

	case 's':
	    data[i++] = (char) 26;
	    break;

	case 'm':
	    data[i++] = (char) 27;
	    break;

	case 'v':
	    data[i++] = (char) 28;
	    break;

	case 'h':
	    data[i++] = (char) 29;
	    break;

	case 'd':
	    data[i++] = (char) 30;
	    break;

	case 'b':
	    data[i++] = (char) 31;
	    break;

	case ' ':
	case '\t':
	case '\n':
	case '\325':
	  break;
	case '>':
        case '#':
        case ':':
	    /* --[A] chunk name / size update */
	    if (j > 0) {
	      /* update previous length if needed */
	      if (i > cumulength)
		chunksize[j-1] = i - cumulength;
	      else
		/* otherwise, no data between : rewrite the new chunk */
		j--;
	    } else {
	      /* j == 0  this is the first chunk */
	      /* "incorrect file format" : remove the "dust" before the first chunk */
	      i = 0;
	    }
	    word = 0xfc000000;
	    cumulength = i;
	    chunkstart[j] = i;
	    chunkname[j] = (char *) MALLOC((unsigned) 256 * sizeof(char));
	    {
	      int c = 0;
	      do {
		/* seek eol/eof */
		car = fgetc(file);
		if (car == ' ' || car == '\t' )
		  car = '_';
		if (car >= 0x20 && car != EOF && car != '\n' && car < 128 && c <= 256)
		  (chunkname[j])[c++] = car;
	      } while (car >= 0x20 && car != EOF && car != '\n' && car < 128);
	      (chunkname[j])[c] = '\0';
	    }
	    j++;
	    break;
	default:
	  /* ignore */;
	} /* switch */
    } /* while */
    fclose(file);

    /* --[B] last chunk size updated */
    if (j > 0) {
      if (i > cumulength)
	chunksize[j - 1] = i - cumulength;
      else
	/* otherwise, no data between : rewrite the new chunk */
	j--;
    }
    
    nbchunks = j;
    
    /* --[C] Raw format : simulate one big fasta chunk */
    if (nbchunks <= 0) {
	nbchunks  = 1;
	chunkname = (char **) MALLOC((unsigned) nbchunks* sizeof(char *));
	ASSERT(chunkname, CreateData);
	chunksize = (int *) MALLOC((unsigned) nbchunks* sizeof(int));
	ASSERT(chunksize, Createdata);
	chunkstart = (int *) MALLOC((unsigned) nbchunks* sizeof(int));
	ASSERT(chunkstart, Createdata);
	chunkname[0] = filename;
	chunksize[0] = datasize;
	chunkstart[0] = 0;
    }

#ifdef DEBUG_CREATEDATA
    for (i = 0; i < nbchunks; i++)
	fprintf(stderr,
		" chunksize = %i chunkstart = %i chunkname = \"%s\" \n",
		chunksize[i], chunkstart[i], chunkname[i]);
#endif

    /* (4) */
    /* FIXME : set some low complexity filter here ... */



    /* (5) return allocated and computed data */
    *p_datasize = datasize;
    *p_data = data;
    if (p_nbchunks != NULL)
	*p_nbchunks = nbchunks;
    if (p_chunkname != NULL)
	*p_chunkname = chunkname;
    if (p_chunksize != NULL)
	*p_chunksize = chunksize;
    if (p_chunkstart != NULL)
	*p_chunkstart = chunkstart;
    return 0;
}




/*
 *   Build a DNA bases of the complementry reverse straind 
 * 
 * - Input  : "data_in"     : DNA array (returned by CreateData for example)
 *            "datasize"    : DNA array size
 * - Output : "p_data_out"  : DNA array containing the complementary reverse straind
 *            
 */


int CreateReverseComplement( /*in */ char *data_in, /*in */ int datasize,
			    /*out */ char **p_data_out)
{
    int i;
    char *data = (char *) MALLOC((unsigned) datasize* sizeof(char));
    ASSERT(data, CreateReverseComplement);

    for (i = 0; i < datasize; i++)
      data[datasize - 1 - i] = (char)COMPLEMENT(data_in[i]);
    
    *p_data_out = data;
    return 0;
}




/*
 * seed pattern size computation 
 * (gp_seed_bitweight, gp_seed_span, gp_seed_bitmask)
 */


int ComputeLength()
{
  int seed;
  int i;
  gp_seeds_span_max = 0;
  gp_seeds_span_min = INFINITY_INT;
  for (seed = 0; seed < gp_nb_seeds; seed++) {
    gp_seeds_bitmask[seed] = 0;
    gp_seeds_bitweight[seed] = 0;
    gp_seeds_span[seed] = 0;
    
    for (i = 0; i < (int) strlen(gp_motifs[seed]); i++) {
      if (gp_motifs[seed][i] == '#') {
	gp_seeds_bitmask[seed]    |= 1;
	gp_seeds_bitweight[seed]  += 2;
      } else {
	if (gp_motifs[seed][i] == '@') {
	  gp_seeds_bitmask[seed]    |= 1;
	  gp_seeds_bitweight[seed]  += 1;
	} else {
	  /* if another char is given : replace it by _*/
	  gp_motifs[seed][i] = '_';
	}
      }
      gp_seeds_span[seed]++;
      gp_seeds_bitmask[seed]<<=1;
    }
       
    gp_motifs[seed][gp_seeds_span[seed]] = '\0';
    gp_seeds_span_max = MAX(gp_seeds_span_max,gp_seeds_span[seed]);
    gp_seeds_span_min = MIN(gp_seeds_span_min,gp_seeds_span[seed]);
  }
  return 0;
}




void AutomataOverlaps(int delph)
{
    int i, j;
    int p_delph = 1 << delph;

    gp_automata = (int *) MALLOC(sizeof(int) * p_delph);
    ASSERT(gp_automata, AutomataOverlaps);
    memset(gp_automata, 0, sizeof(int) * p_delph);

    for (i = 0; i < p_delph; i++) {
	j = i;
	while (j != 0) {
	    j = j & (j - 1);
	    gp_automata[i]++;
	}
    }
}


static void BuildRedMask(int seed, int * first) {  
  int  mask_copy  = 0;
  int  c_log      = 0;
  /* precompute some seed positions */
  {
    int i;
    for ( i = 0 ; i < gp_seeds_span[seed] ; i++ ) {
      char seed_element = gp_motifs[seed][i];
      switch (seed_element) {
      case '#' :
	mask_copy |= 1;
	c_log++;
	mask_copy <<= 1;
	break;
      case '@' :
	c_log++;
	mask_copy <<= 1;
	break;
      }
    }
  }
  
  /* for all pair of letters */ 
  {
    int  a; 
    for(a = 0 ; a < 4 ; a++ ) {
      int b;
      for(b = 0 ; b <= a ; b++ ) {
	int c;
	
	/* build all words containing only 'a' and 'b' letters */
	for(c = 0 ; c < (1 << c_log); c++){
	  int keycode = 0;
	  int i = 0;
	  
	  /* build (a,b)-key code */
	  for( i = 0 ; i < c_log ; i++){
	    int i_pow = 1<<i;
	    if (i_pow & mask_copy){
	      keycode <<= 2;
	      keycode |=  ((c&i_pow)?a:b);
	    }else{
	      keycode <<= 1;
	      keycode |=  ((c&i_pow)?a:b)&1;
	    }
	  }
	  
	  /* reset this keycode only if weight >= 8 */
	  if (gp_seeds_bitweight[seed] >= 8*2)
	    first[keycode] = -2;
	}
      }
    }
  }
}


/*
 * Create a linkedlist of kwords with param "list" and index "first"
 *
 * - Input : "data"    : dna input table 
 *           "datasize": dna input table size
 *           "seed"    : seed number
 * - Output: "p_first" : return a table of size (2^w). For each code w, it contains the first 
 *                       position occurence of the seed with such code on the sequence
 *           "p_list"  : return a table of size (keysize). For each position i, it contains next
 *                       position j such that seed i and j code are equal. 
 */


#define ISTART 0

int CreateKeyList( /* in */ char *data,     /* in  */ int datasize,    /* in */ int seed,
		   /* out */ int **p_first, /* out */ int * p_first_size, 
		   /* out */ int **p_list,  /* out */ int * p_list_size, /* out */ Feature * feature)
{


  int *bucklet = NULL;
  int *first = NULL;
  int *list = NULL;
  int *count = NULL;
#ifdef KEYLISTCOMPRESS
  int *listcompress = NULL;
#endif
  int sizeofbucklet = 1;
  int keysize = datasize - gp_seeds_span[seed] + 1;
    
  /* prune algorithm */
  int maxchainsize = 0;
  int nbkeys       = 0;
  int fastcountall = 0;
  int fastcounter[sizeof(unsigned int) * 8] = {0};
  int imask      [sizeof(unsigned int) * 8] = {0};
  
  int  i=0, j=0, w=0;

  /* (1) allocation of 3 table */
  sizeofbucklet <<= (gp_seeds_bitweight[seed]);
  
  bucklet = (int *) MALLOC((unsigned) sizeofbucklet * sizeof(int));
  ASSERT(bucklet, CreateKeyList);
  memset(bucklet, 0x00, (unsigned) sizeofbucklet * sizeof(int));
  
  first = (int *) MALLOC((unsigned) (sizeofbucklet+1) * sizeof(int));
  ASSERT(first, CreateKeyList);
  memset(first, 0xff, (unsigned) sizeofbucklet * sizeof(int));
  
  count = (int *) MALLOC((unsigned) sizeofbucklet * sizeof(int));
  ASSERT(count, CreateKeyList);
  memset(count, 0x0, (unsigned) sizeofbucklet * sizeof(int));
  
  list  = (int *) MALLOC((unsigned) keysize * sizeof(int));
  ASSERT(list, CreateKeyList);
  memset(list, 0xff, (unsigned) keysize * sizeof(int));
  
  /* (2) list creation algorithm */
  BuildRedMask(seed,first); /* 
			     * redmask (-2) is for unallowed key = "0, all G, all C, all A, all T" 
			     * and dinucleotides only composed keys ....
			     */
  for (j = 0; j < keysize; j++) {
    KEY(w,data,j,seed);
    if ( first[w] != -2 ) {
      if ( first[w] == -1 ) {
	first[w] = j;
      } else {
	list[bucklet[w]] = j;
      }
      bucklet[w] = j;
      count[w]++;
      nbkeys++;
      maxchainsize = MAX(count[w], maxchainsize);
    }
  }
  FREE(bucklet,(unsigned) sizeofbucklet * sizeof(int));
  
  
  /* (3) median-mean selection */
  
  /* (3.0) init counters */
  imask[0] = 1;
  for (i = 1 ; i < sizeof(int)*8 ; i++ ) {
    imask[i] = ((imask[i-1]) << 1) | 1 ;
    fastcounter[i] = 0;
  }
  
  /* (3.1) fast approximate counting */ 
  for (w = 0; w < sizeofbucklet; w++) {
    for (i = ISTART ; i < sizeof(int)*8 ; i++ ){
      if ( ( imask[i]  & count[w] ) == count[w]) {
	fastcounter[i]++;
	break;
      }
    }
  }

#ifdef DEBUGFASTCOUNT
  for  (i = ISTART ; i < sizeof(int)*8 ; i++ ) {
    fprintf(stderr, " fastcount[%d] = %d \n", i , fastcounter[i]);
  }
#endif

  /* (3.2) select part to be removed part */
  for (i = ISTART ; i < sizeof(int)*8 ; i++ ) {
    fastcountall += fastcounter[i];
    if (fastcountall * 1.0 > sizeofbucklet * (1.0 - (gp_t + 1e-32)/sqrt(sizeofbucklet)))
      break;
  }

  /* (3.3) remove gp_t percent */
  for (w = 0; w < sizeofbucklet; w++) {
    /* normalize */
    if (first[w] < 0){
      first[w] = -1;
    }else {
      if ((count[w] & imask[i]) != count[w]) {
	first[w] = -1;
	nbkeys -= count[w];
#ifdef STATS	
	feature->nb_keys_removed++;
#endif
#ifdef DEBUGRMKEY
	fprintf(stderr,"- removing key : w=%x, nbocc=%d\n",w,count[w]); 
#endif
      }
    }
  }
  
  FREE(count,(unsigned) sizeofbucklet * sizeof(int));
  
  /* (4) keylist compression */
#ifdef KEYLISTCOMPRESS
  listcompress =  (int *) MALLOC((unsigned) (nbkeys + 1) * sizeof(int));
  ASSERT(listcompress, CreateKeyList);
  i = 0;
  for (w = 0; w < sizeofbucklet; w++) {
    j = first[w];
    first[w] = i;
    while (j >= 0) {
      listcompress[i++] = j;
      j = list[j];
    }
  }
  first[sizeofbucklet] = i;
#endif

  /* (5) list "copy" */
#ifdef KEYLISTCOMPRESS
  FREE(list, keysize * sizeof(int));
  *p_list       = listcompress;
  *p_list_size  = (nbkeys  + 1 ) * sizeof(int);
#else
  *p_list       = list;
  *p_list_size  = keysize * sizeof(int);
#endif
  *p_first      = first;
  *p_first_size = (sizeofbucklet + 1) * sizeof(int);
  return 0;

}




/*
 * COMPLEXITY CHECK
 */
double Entropy(char *data, int pos, int len, int k)
{
    int p4k = 1 << (k << 1);	/*4^k */
    int mask = 0;
    int key = 0;
    int i = 0;
    double result = 0.0;
    int *freq = (int *) MALLOC(p4k * sizeof(int));
    ASSERT(freq, Entropy);
    memset(freq,'\0',sizeof(int)*p4k);
    
    for (i = 0; i < k; i++) {
	mask <<= 2;
	mask |= 0x3;
	key <<= 2;
	key |= backcode[(int) data[pos + i]];
    }


    for (i = k; i < len; i++) {
	key <<= 2;
	key |= backcode[(int) data[pos + i]];
	key &= mask;
	freq[key]++;
    }


    for (i = 0; i < p4k; i++) {
	if (freq[i] != 0) {
	    result -=
		((double) (freq[i]) / (double) (len - k + 1)) *
		log((double) (freq[i]) / (double) (len - k + 1));
	}
    }
    result /= log(2.0000);
    FREE(freq,p4k * sizeof(int));
    return result;
}




/*
 * Display sequence fragments
 */

int DisplaySequence(char *data, int pos, int len) {
  
    int j;
    for (j = pos; j < pos + len; j++) {
      fprintf(stdout,"%c", lookup[(int) data[j]]);
    }
    fprintf(stdout,"\n");
    return 0;
}


int DisplayData(char * filename, char * data, int size, int nbchunks, char ** chunkname, int * chunksize, int * chunkstrt) {
  int i;
  fprintf(stderr,"filename : \"%s\"\n",filename);
  fprintf(stderr,"- size : %d\n",size);
  fprintf(stderr,"- nbchunks : %d\n",nbchunks);
  for( i = 0 ; i < nbchunks ; i++ ) {
    fprintf(stdout,">\"%s\"|[start:%d,size:%d]\n",chunkname[i],chunkstrt[i],chunksize[i]);
    DisplaySequence(data,chunkstrt[i],chunksize[i]);
  }
  return 0;
}

int DisplaySeeds(){
  int i = 0;
  for( i = 0 ; i < gp_nb_seeds ; i++){
    fprintf(stderr,"- gp_motif[%d] = \"%s\", span = %3d , bitweight = %3d\n",i,gp_motifs[i],gp_seeds_span[i],gp_seeds_bitweight[i]);
  }
  return 0;
}
