/*
 *  YASS 1.11
 *  Copyright (C) 2004-2007
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
/* #include <unistd.h> */
#include <string.h>
#include "global_var.h"
#include "tuple.h"
#include "prdyn.h"

#define  CHAR_EQU '|'
#define  CHAR_SS  ':'
#define  CHAR_SV  '.'
#define  CHAR_DEL ' '
#define  CHAR_INS ' '
#define  CHAR_DUM '-'

#define GAP1 (-2)  /* gap cost */
#define SUB1 (-1)  /* substitution cost */

#define TNORM(a) (((a)%16) == 7?(3):((a)%16))

/*
 *  [I]
 *
 *   "quick" alignement Functions used in the alignment part
 *
 ***********************************************************************/



/*
 * Run this method before any use of "quick" alignment algorithm
 * (it allocate dynamic programming table)
 *
 */

#ifdef INLINE
 inline 
#endif
int initialise_alignment(int size,
			 Feature *feature)
{
  feature->buffer01 = (int *)MALLOC((unsigned)(size+1)* sizeof(int));
  ASSERT(feature->buffer01,initialise_aligment);
  return 0;
}


/*
 * Align two segments defined by their started pos and their length 
 */

#ifdef INLINE
 inline 
#endif
int alignment(char * data1, int pos1, int len1,
	      char * data2, int pos2, int len2,
	      Feature *feature){
  
  int i,j;
  int tmp_diag,tmp_diag_prev;
  
  for(i=0;i<len1+1; i++)
    feature->buffer01[i] = i*GAP1;
  
  /* columns */
  for(j=1;j<len2+1;j++){
    
    tmp_diag_prev = (j-1)*GAP1;
    feature->buffer01[0] = j*GAP1;
    
    
    /* lines*/
    for(i=1;i<len1+1;i++){
      tmp_diag = feature->buffer01[i];

      /* compute value */
      if (TNORM(data2[pos2 + j-1]) == TNORM(data1[pos1 + i-1]))
	feature->buffer01[i] = MAX3(tmp_diag_prev,     feature->buffer01[i]+GAP1, feature->buffer01[i-1]+GAP1);
      else
	feature->buffer01[i] = MAX3(tmp_diag_prev + SUB1, feature->buffer01[i]+GAP1, feature->buffer01[i-1]+GAP1);
      
      /* update diag value */
      tmp_diag_prev = tmp_diag;
    }
  } 
  return feature->buffer01[len1];
}

/*
 * Try to align two subparts;
 * return -1 si failed ( "leavethreashold" is over "steps_to_leave") or 
 * the result (then /out/ positions are filled)
 */

#ifdef INLINE
 inline 
#endif
int left_alignment(char * data1, int pos1, int len1,
		   char * data2, int pos2, int len2, 
		   double leavethreshold,int steps_to_leave,
		   /*out*/ int * p_gain_score,/*out*/ int * p_last_pos1,/*out*/ int * p_last_pos2,
		   Feature *feature)
{
  int i,j;
  int tmp_diag,tmp_diag_prev;
  int steps_over_threshold = 0;
  int max_score_in_column = 0;
  int i_max = 0;
  

  for(i=0;i<len1+1; i++)
    feature->buffer01[i] = i*GAP1;
  
  /* columns */
  for(j=1;j<len2+1;j++){
    
    tmp_diag_prev        = (j-1)*GAP1;
    feature->buffer01[0] =    j *GAP1;
    
    max_score_in_column = -INFINITY_INT;
    
    
    /* lines */
    for(i=1;i<len1+1;i++){
      tmp_diag = feature->buffer01[i];

      /* compute value */
      if (TNORM(data2[pos2 + j-1]) == TNORM(data1[pos1 + i-1]))
	feature->buffer01[i] = MAX3(tmp_diag_prev,     feature->buffer01[i]+GAP1, feature->buffer01[i-1]+GAP1);
      else
	feature->buffer01[i] = MAX3(tmp_diag_prev + SUB1, feature->buffer01[i]+GAP1, feature->buffer01[i-1]+GAP1);
       
      if (feature->buffer01[i] > max_score_in_column){ 
	max_score_in_column = feature->buffer01[i]; 
	i_max = i;
      }
      
      /* update diag value */
      tmp_diag_prev = tmp_diag;
    }

    if (((double)(max_score_in_column))/((double)j) < leavethreshold )
      steps_over_threshold++;
    if (steps_over_threshold > steps_to_leave){
      *(p_last_pos1)  = pos1 - i_max;
      *(p_last_pos2)  = pos2 - j;
      *(p_gain_score) = max_score_in_column;
      return -1;
    }
  }/* for */

  *(p_last_pos1)  = pos1 - i_max;
  *(p_last_pos2)  = pos2 - j;
  *(p_gain_score) = max_score_in_column;

  return feature->buffer01[len1];
}



/*
 * Try to align two subparts;
 * return -1 si failed ( "leavethreashold" is over "steps_to_leave") or 
 * the result (then /out/ positions are filled)
 */

#ifdef INLINE
 inline 
#endif
 int right_alignment(char * data1, int pos1, int len1,
		     char * data2, int pos2, int len2, 
		     double leavethreshold,int steps_to_leave,
		     /*out*/ int * p_gain_score,/*out*/ int * p_last_pos1,/*out*/ int * p_last_pos2,Feature *feature)
{
  
  int i,j;
  int tmp_diag,tmp_diag_prev;
  int steps_over_threshold = 0;
  int max_score_in_column = 0;
  int i_max = 0;
  

  for(i=0;i<len1+1; i++)
    feature->buffer01[i] = i*GAP1;
  
  /* columns */
  for(j=1;j<len2+1;j++){
    
    tmp_diag_prev = (j-1)*GAP1;
    feature->buffer01[0] = j*GAP1;
    
    max_score_in_column = -INFINITY_INT;
    
    
    /* lines */
    for(i=1;i<len1+1;i++){
      tmp_diag = feature->buffer01[i];

      /* compute value */
      if (TNORM(data2[pos2 - j+1]) == TNORM(data1[pos1 - i+1]))
	feature->buffer01[i] = MAX3(tmp_diag_prev,     feature->buffer01[i]+GAP1, feature->buffer01[i-1]+GAP1);
      else
	feature->buffer01[i] = MAX3(tmp_diag_prev + SUB1, feature->buffer01[i]+GAP1, feature->buffer01[i-1]+GAP1);
      
      if (feature->buffer01[i] > max_score_in_column){ 
	max_score_in_column = feature->buffer01[i]; 
	i_max = i;
      }
     
      /* update diag value */
      tmp_diag_prev = tmp_diag;
    }
    
    if (((double)(max_score_in_column))/((double)j) < leavethreshold )
      steps_over_threshold++;
    if (steps_over_threshold > steps_to_leave){
      *(p_last_pos1) = pos1 + i_max;
      *(p_last_pos2) = pos2 + j;
      *(p_gain_score) = max_score_in_column;
      return -1;  
    }
  }/* for */

  *(p_last_pos1) = pos1 + i_max;;
  *(p_last_pos2) = pos2 + j;
  *(p_gain_score) = max_score_in_column;

  return feature->buffer01[len1];
}




#define I2(i,j) (feature->buffer01[(i)+ (((j)&1)?(len1+1):(0))])
#define M2(i,j) (feature->buffer01[(i)+ (((j)&1)?(len1+1):(0)) + ((len1+1)*2)])
#define S2(i,j) (gp_substitution_matrix[(int)(data1[pos1+(i)-1])][ ((int)(data2[pos2+(j)-1]))])

/*
 * Compute alignment score with :
 * - gap cost 
 * - gap extend
 * - subs 4x4 
 */

#ifdef INLINE
 inline 
#endif
 int alignment_SG(char * data1, int pos1, int len1,
		  char * data2, int pos2, int len2,
		  Feature *feature)
{
  int i,j;
  
  M2(0,0) = 0;
  I2(0,0) = gp_cost_gap_opened -  gp_cost_gap_continued;
  
  for(i=1;i<len1+1; i++)
    M2(i,0) = -INFINITY_INT;
  
  for(i=1 ; i<len1+1 ; i++)
    I2(i,0) = I2(i-1,0) +  gp_cost_gap_continued;
  
  for(j=1 ; j<len2+1 ; j++){
    
    M2(0,j) =  -INFINITY_INT;
    I2(0,j) =  I2(0,j-1) + gp_cost_gap_continued;
    
    for(i=1;i<len1+1;i++){
      I2(i,j) = MAX (
		     MAX( M2(i,j-1) + gp_cost_gap_opened    , M2(i-1,j) + gp_cost_gap_opened )
		     ,
		     MAX( I2(i,j-1) + gp_cost_gap_continued ,  I2(i-1,j) + gp_cost_gap_continued )
		     );
      M2(i,j) = MAX ( M2(i-1,j-1) + S2(i,j), I2(i-1,j-1) + S2(i,j) );
    }
  }
  return MAX(M2(len1,len2),I2(len1,len2));
}

#ifdef INLINE
 inline 
#endif
 
 int alignment_SG_Strait(char * data1, int pos1, char * data2, int pos2, int len,
			 Feature *feature)
{
  int i,result = 0;
  for(i=0; i < len ;i++){
    result += gp_substitution_matrix[(int)(data2[pos2+i])][(int)data1[pos1+i]];
#ifdef DEBUG_ALIGNMENT_STRAIT
    fprintf(stdout,"* score[%c/%c]=%d\n",lookup[data2[pos2+i]],lookup[data1[pos1+i]],gp_substitution_matrix[(int)(data2[pos2+i])][(int)data1[pos1+i]]); 
#endif
  }
  return result;
}



#ifdef INLINE
 inline 
#endif
 int alignment_SG_on_MA(char * data1,  char * data2,
			MA * ma, int left_correction,
			Feature *feature)
{
  int i,result = 0;
  int left_size,right_size;
  tuple * t = ma->first_tuple, * t_prev = ma->first_tuple ;
  
  /* [1] compute score before first tuple */
  
  left_size  = TBL_POS(t) - ma->left_pos_begin  + 1 ; /*  + 0 */
  right_size = TBR_POS(t) - ma->right_pos_begin + 1 ; /*  + 0 */
  
  if ( left_size > 0 && right_size > 0)
    result += alignment_SG(data1, ma->left_pos_begin , left_size ,
			   data2, ma->right_pos_begin, right_size,
			   feature
			   );
  
  /* [2] compute score for several tuples */
  while(t != NULL){
    /* (1) compute score inside tuple */
    for(i = 0 ;i < TSIZE(t) ; i++)
      result += gp_substitution_matrix[(int)(data1[i+TBL_POS(t)])][(int)(data2[i+TBR_POS(t)])];
	
    if (t->next){
      /* (2) comptue score between tuple (t) and (t->next) */
      result += alignment_SG(data1,TEL_POS(t),TGAP_L(t,t->next),
			     data2,TER_POS(t),TGAP_R(t,t->next),
			     feature
			     );
    }/* if */
    t_prev = t;
    t = t->next;
  }/* while */

  /* [3] compute score after last tuple */
  
  left_size  =  ma->left_pos_end  - TEL_POS(t_prev) + 1;
  right_size =  ma->right_pos_end - TER_POS(t_prev) + 1;


  if ( left_size > 0 && right_size> 0)
     result += alignment_SG(data1, TEL_POS(t_prev), left_size ,
			    data2, TER_POS(t_prev), right_size,
			    feature
			    );
  
  return result;
}


#ifdef INLINE
 inline 
#endif
 int alignment_SG_DROP(char * data1, int pos1, int len1,
		       char * data2, int pos2, int len2,
		       int XDROP,
		       Feature *feature
		       )
{
  int i,j;
  int max;
  
  M2(0,0) = 0;
  I2(0,0) = gp_cost_gap_opened -  gp_cost_gap_continued;
  
  for(i=1;i<len1+1; i++)
    M2(i,0) = -INFINITY_INT;
  
  for(i=1 ; i<len1+1 ; i++)
    I2(i,0) = I2(i-1,0) +  gp_cost_gap_continued;
  
  for(j=1 ; j<len2+1 ; j++){
    
    M2(0,j) =  -INFINITY_INT;
    I2(0,j) =  I2(0,j-1)  + gp_cost_gap_continued;
    max     =  M2(0,j);
    
    for(i=1;i<len1+1;i++){
      I2(i,j) = MAX (
		     MAX( M2(i,j-1) + gp_cost_gap_opened    , M2(i-1,j) + gp_cost_gap_opened )
		     ,
		     MAX( I2(i,j-1) + gp_cost_gap_continued ,  I2(i-1,j) + gp_cost_gap_continued )
		     );
      M2(i,j) = MAX ( M2(i-1,j-1) + S2(i,j), I2(i-1,j-1) + S2(i,j) ); 
      max = MAX(max,M2(i,j));
    }
    if (max < -XDROP)
      return max;
  }
  return MAX(M2(len1,len2),I2(len1,len2));
}


#define M2opt_current(a)   (*(feature->buffer01 + d_current + (a)))
#define M2opt_previous(a)  (*(feature->buffer01 + d_previous + (a)))
#define I2opt_current(a)   (*(feature->buffer01 + I2_DELTA + d_current + (a))) 
#define I2opt_previous(a)  (*(feature->buffer01 + I2_DELTA + d_previous + (a)))
#define S2opt(i,j) (gp_substitution_matrix[(int)(data1[pos1+(i)-1])][(int)(data2[pos2+(j)-1])])

#ifdef INLINE
 inline 
#endif
 int alignment_SG_DROP_opt(char * data1, int pos1, int len1,
			   char * data2, int pos2, int len2,
			   int XDROP,
			   Feature *feature)
{
  int i,j,s;
  int max;
  int I2_DELTA = (len1 + 1)<<1;
  int DELTA    = (len1 + 1);
  int d_previous = DELTA;
  int d_current  = 0;

  M2opt_current(0) = 0;
  I2opt_current(0) = gp_cost_gap_opened -  gp_cost_gap_continued;
  
  for(i=1;i<len1+1; i++){    
    M2opt_current(i) = -INFINITY_INT;
    I2opt_current(i) = I2opt_current(i-1) +  gp_cost_gap_continued;
  }
  

  for(j=1 ; j<len2+1 ; j++){
    if (j&1){
      d_previous = 0;
      d_current  = DELTA;
    }else{
      d_previous = DELTA;
      d_current  = 0;
    }
    
    
    M2opt_current(0) =  -INFINITY_INT;
    I2opt_current(0) =  I2opt_previous(0)  + gp_cost_gap_continued;
    max              =  -INFINITY_INT;
    
    for(i=1;i<len1+1;i++){ 
      d_previous ++;
      d_current  ++;
      I2opt_current(0) = MAX (
			      MAX( M2opt_previous(0) + gp_cost_gap_opened    , M2opt_current(-1) + gp_cost_gap_opened )
			      ,
			      MAX( I2opt_previous(0) + gp_cost_gap_continued , I2opt_current(-1) + gp_cost_gap_continued )
			      );
      s = S2opt(i,j);
      M2opt_current(0) = MAX ( M2opt_previous(-1) + s, I2opt_previous(-1) + s ); 
      max = MAX (max,M2opt_current(0));
    }
    if (max < -XDROP)
      return max;
  }
  return MAX(M2opt_current(0),I2opt_current(0));
}





#define D_I2(k,i) (feature->buffer01[(i) + ((k)&1?(MIN(len1,len2)+1):0) ])
#define D_M2(k,i) (feature->buffer01[(i) + ((k)&1?(MIN(len1,len2)+1):0) + 2*(MIN(len1,len2)+1) ])
#define D_S2(i,j) (gp_substitution_matrix[(int)(data1[pos1+(i)])][(int)(data2[pos2+(j)])])


#ifdef INLINE
 inline 
#endif
 int  alignment_SG_Lz(char * data1, int pos1, int len1, 
		      char * data2, int pos2, int len2, 
		      int XDROP,
		      Feature *feature)
{
  
  int last_score  = 0, score  = 0;
  int last_indels = 0, indels = 0;
  int last_i      = 0, i      = 0;

  if (len1 < len2){
    
    for(i=0 ; i < len1 ; i++){
      score += D_S2(i,i+indels);
      if (score > last_score){
	last_score  = score;
	last_i      = i;
	last_indels = indels;
      }
      
      if (score < -XDROP){
	score  = last_score + gp_cost_gap_opened + ABS(indels-last_indels) * gp_cost_gap_continued;
	i      = last_i;
	indels++;
	if (indels > len2 - len1)
	  return last_score;
      }
    }/* for i */
    return last_score;
  
  } else {
     
    for(i=0 ; i < len2 ; i++){
      score += D_S2(i+indels,i);
      if (score > last_score){
	last_score  = score;
       	last_i      = i;
	last_indels = indels;
      }
      
      if (score < -XDROP){
	score  = last_score + gp_cost_gap_opened + ABS(indels-last_indels) * gp_cost_gap_continued;
	i      = last_i;
	indels++;
	if (indels > len1 - len2)
	  return last_score;
      }
    }/* for i */
    return last_score;
  }
  
}






#define INVERSE(BUF1,BUF2,BUFT) BUFT=BUF1;BUF1=BUF2;BUF2=BUFT;

/***********************************
  Exemple d'une recherche de score
  par la methode border I={16,4}

Tableau des M: Bm0 represente la borne a l'extermier du parcourt du tableau
               Bm1 est la borne suivant


      T   G   A  C   G
   0  16  20  24 28  32
A  16  4  17  17 28  29   30 
C  20 20   5  21 15  25   12  40
G  24 17  24  9  6   19   26  30  
G  28 25  21  25 29  7    55  51
T  32 32  26  18 22  [23]  ...
      30  34  17 25   ...
          25  32 36   ...
le milieu est represent pas width


Tableau des I: permet de savoir le coups minimum d'un GAP


************************************/

int alignment_SG_Border2 (char * data1, int pos1, int len1, 
			 char * data2, int pos2, int len2,
			 int XDROP,
			 Feature *feature){
  /* band alignment length anw width*/
  int dwidth       = len2-len1;
  int bandwidth    = gp_delta_stat;

  /* various switch elements to access information */
  int row1       = 2*bandwidth + dwidth;
  int row2       = 2*row1;

  /* dp tables */
  int * bm0 = (int *)feature->buffer01;
  int * bm1 = (int *)feature->buffer01+row1;
  int * bi0 = (int *)feature->buffer01+row2;
  int * bi1 = (int *)feature->buffer01+row2+row1;
  int * tmp;


  /* maximal and minimal positions taken into account */
  int j,k;
  int jmin = 0, jmax = row1;
  int i;


  pos1++;pos2++; /* Del the first tuple */
  len1-=2;len2-=2;
  
  /* init row*/
  for(j=0;j<row1;j++)
    bm0[j] = bi0[j] = (gp_cost_gap_opened) + 
      (ABS(j-bandwidth)-1)*gp_cost_gap_continued;  
  bm0[bandwidth] = bi0[bandwidth] = 0;
  /*fprintf(stderr,"len1=%d, len2=%d\n",len1,len2);*/


  if (len2<len1) {
    char * tmp_char;
    int  tmp_int;
    /* Invertion des sequences*/
    INVERSE(data1,data2,tmp_char);
    INVERSE(pos1 ,pos2 ,tmp_int);
    INVERSE(len1 ,len2 ,tmp_int);
  }

  if (len1 <= len2){
    for(k=0 ; k< len1; k++){
      int d;

      /* Methode horizontale */
      i=1;
      for (j=bandwidth-1;j>jmin;j--) {
	d=gp_substitution_matrix[(int)(data2 + pos2)[k+i]][(int)((data1 + pos1)[k])];      
	bm1[j] = MAX( bm0[j] , bi0[j] ) + d;
	i++;
      }
      /* Methode verticale */
      i=0;
      for (j=bandwidth;j<jmax;j++) {
	d=gp_substitution_matrix[(int)(data2 + pos2)[k]][(int)((data1 + pos1)[k+i])];     
	bm1[j] = MAX( bm0[j] , bi0[j] ) + d;
	i++;
      }

      /* MinxMax evaluations */
      j=jmin;	
      while( bm1[j++] < -XDROP && j<jmax)
	jmin++;
      j=jmax;
      while( bm1[j--] < -XDROP && j>0)
	jmax--;
      
      if (jmax<jmin)
	return -XDROP-1;
#ifdef DEBUG_ALIGNMENT_BORDER
      fprintf(stderr,"Bm0:");
      for (j=k;j<=jmax-k  ;j++) {
	fprintf(stderr,"%d\t",bm0[j]);
      }
      fprintf(stderr,"\nBm1:");
      for (j=k+1;j<jmax-k  ;j++) {
	fprintf(stderr,"%d\t",bm1[j]);
      }
      fprintf(stderr,"\n");
#endif
      bi1[bandwidth] = MAX(
			   MAX(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , 
			   MAX(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened);
     
      for( j=bandwidth-1; j>jmin ; j--){
	bi1[j] = MAX(
		     MAX(bm0[j],   bm1[j+1] ) + gp_cost_gap_opened,
		     MAX(bi0[j],   bi1[j+1] ) + gp_cost_gap_continued
		     );
      }
      for( j=bandwidth+1; j<jmax ; j++){
	bi1[j] = MAX(
		     MAX(bm0[j+1],   bm1[j-1] ) + gp_cost_gap_opened,
		     MAX(bi0[j+1],   bi1[j-1] ) + gp_cost_gap_continued
		     );
      }
#ifdef DEBUG_ALIGNMENT_BORDER
      fprintf(stderr,"Bi0  ");
      for (j=jmin+k;j<=jmax  ;j++) {
	fprintf(stderr,"%d\t",bi0[j]);
      }
      fprintf(stderr,"\nBi1  ");
      for (j=jmin+k;j<jmax  ;j++) {
	fprintf(stderr,"%d\t",bi1[j]);
      }
      fprintf(stderr,"\n");
#endif  
      /*Interchange */
      INVERSE(bm0,bm1,tmp);
      INVERSE(bi0,bi1,tmp);
    }
  }
#ifdef DEBUG_ALIGNMENT_BORDER
  for (j=0;j<jmax  ;j++) {
    fprintf(stderr,"%d\t",bm0[j]);
  }
  fprintf(stderr,"\n");
  fprintf (stderr,"\nband=%d,dwidth=%d\n",bandwidth, dwidth);
  fprintf (stderr,"\nM:%d,%d\n",bm0[bandwidth + dwidth],gp_cost_match);
#endif
  return MAX(bm0[bandwidth + dwidth],bi0[bandwidth + dwidth]);/* del the last match*/

}






/*
 * band alignment
 *
 *
 *          ------------------
 *   	    |^               |
 *          | bandwidth      |
 *          |v               |
 *          |----------------|	
 *          |I dlen          |
 *          |----------------|
 *          |^               |
 *          | bandwidth      |
 *          |v               |
 *	    ------------------
 *
 *
 *
 *  Classical DP matrix :
 *       (i) ->
 *     1  2  4  7  11 16 22
 * (j) 3  5  8  12 17 23
 *  |  6  9  13 18 24
 *  v  10 14 19 25
 *     15 20 26
 *     21 27
 *     28
 *
 *
 *
 *  Border DP matrix
 *  
 *               (k)->
 *     22
 *     16----------        (i,j-1)
 *(i)  11 23                     \
 * |   7  17            (i-1,j-1)-[*]
 * V   4  12 24                    |                   (i,j-1)
 *     2  8  18                  (i-1,j)                    \
 *---- 1  5  13  25 ----------------------------- (i-1,j-1)-[*]
 *     3  9  19                  (i,j-1)                    /
 *     6  14 26                    |                   (i-1,j)
 *     10 20            (i-1,j-1)-[*]
 *     15 27                     /
 *     21----------        (i-1,j)
 *     28
 */

int alignment_SG_Border(char * data1, int pos1, int len1, 
			char * data2, int pos2, int len2,
			int XDROP,
			Feature *feature)
{
  /* band alignment length anw width*/
  int dwidth       = ABS(len2-len1);
  int bandwidth    = gp_delta_stat;
  
 
  /* various switch elements to access information */
  int row1       = 2*bandwidth + dwidth;
  int row2       = 2*row1;
  
  /* dp tables */
  int * bm0 = (int *)feature->buffer01;
  int * bm1 = (int *)feature->buffer01+row1;
  int * bi0 = (int *)feature->buffer01+row2;
  int * bi1 = (int *)feature->buffer01+row2+row1;
  
  /* maximal and minimal positions taken into account */
  int j,k;
  int jmin = 0, jmax = row1;

  for(j=0;j<row1;j++)
    bm0[j] = bi0[j] = (gp_cost_gap_opened) + ABS(j-bandwidth)*gp_cost_gap_continued;
  
  bm0[bandwidth] = bi0[bandwidth] = 0;

  if (len1 < len2){
    
    for(k=1 ; k< len1; k++){ 
     
      if(k&1){	
	/* M(j) full vectorisable loop */
	for(j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)(data2 + pos2)[j+k-bandwidth]][(int)((data1 + pos1)[k])];
	  bm1[j] = MAX( bm0[j] , bi0[j] ) + d; 
	}	
	
	/* MinxMax evaluations */
	j=jmin;	
	while( bm1[j++] < -XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm1[j--] < -XDROP && j>0)
	  jmax--;
	if (jmax<jmin)
	  return -XDROP-1;
  

	/* I(j) in two half loops */
	bi1[bandwidth] = MAX(MAX(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , MAX(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi1[j-1] = MAX(
			 MAX(bi0[j-1],bi1[j]) + gp_cost_gap_continued,
			 MAX(bm0[j-1],bm1[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi1[j] = MAX(
		       MAX(bi0[j],bi1[j-1]) + gp_cost_gap_continued,
		       MAX(bm0[j],bm1[j-1]) + gp_cost_gap_opened
		       );
	}

      }else{/* k&1*/
          
      	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)(data2 + pos2)[j+k-bandwidth]][(int)((data1 + pos1)[k])];
	  bm0[j] = MAX( bm1[j] , bi1[j] ) + d;
	}	

	/* MinxMax evaluations */
	j=jmin;	
	while( bm0[j++] < -XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm0[j--] < -XDROP && j>0)
	  jmax--;
	if (jmax<jmin)
	  return -XDROP-1;
	

	/* I(j) in two half loops */
	bi0[bandwidth] = MAX(MAX(bi1[bandwidth-1],bi1[bandwidth+1]) + gp_cost_gap_continued , MAX(bm1[bandwidth-1],bm1[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi0[j-1] = MAX(
			 MAX(bi1[j-1],bi0[j]) + gp_cost_gap_continued,
			 MAX(bm1[j-1],bm0[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi0[j] = MAX(
		       MAX(bi1[j],bi0[j-1]) + gp_cost_gap_continued,
		       MAX(bm1[j],bm0[j-1]) + gp_cost_gap_opened
		       ); 
	}
      }/* k&1*/
    }/* k */

  }else{
    
    for( k=1 ; k< len2; k++){ 
          
      if(k&1){	
	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data2 + pos2)[k])][(int)(data1 + pos1)[j+k-bandwidth]];
	  bm1[j] = MAX( bm0[j] , bi0[j] ) + d;
	}	

	/* MinxMax evaluations */
	 j=jmin;	
	while( bm1[j++] < -XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm1[j--] < -XDROP && j>0)
	  jmax--;
	
	if (jmax<jmin)
	  return -XDROP-1;

	/* I(j) in two half loops (each on is vectorisable on "non stupid" compilers) */
	bi1[bandwidth] = MAX(MAX(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , MAX(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi1[j-1] = MAX(
			 MAX(bi0[j-1],bi1[j]) + gp_cost_gap_continued,
			 MAX(bm0[j-1],bm1[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi1[j] = MAX(
		       MAX(bi0[j],bi1[j-1]) + gp_cost_gap_continued,
		       MAX(bm0[j],bm1[j-1]) + gp_cost_gap_opened
		       );
	}

      }else{/* k&1*/
          
      	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data2 + pos2)[k])][(int)(data1 + pos1)[j+k-bandwidth]];
	  bm0[j] = 
	    MAX( bm1[j] , 
		 bi1[j] ) 
	    + d;
	}	

	/* MinxMax evaluations */
	j=jmin;
	while( bm0[j++] < -XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm0[j--] < -XDROP && j>0)
	  jmax--;
	
	if (jmax<jmin)
	  return -XDROP-1;
	
	/* I(j) in two half loops (each on is vectorisable on "non stupid" compilers) */
	bi0[bandwidth] = MAX(MAX(bi1[bandwidth-1],bi1[bandwidth+1]) + gp_cost_gap_continued , MAX(bm1[bandwidth-1],bm1[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi0[j-1] = MAX(
			 MAX(bi1[j-1],bi0[j]) + gp_cost_gap_continued,
			 MAX(bm1[j-1],bm0[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi0[j] = MAX(
		       MAX(bi1[j],bi0[j-1]) + gp_cost_gap_continued,
		       MAX(bm1[j],bm0[j-1]) + gp_cost_gap_opened
		       ); 
	}
      }/* k&1*/
    }/* k */
 
  }
  return (MAX(len1,len2)&1)?(bm0[bandwidth + dwidth]):(bm1[bandwidth + dwidth]);
}





#ifdef INLINE
 inline 
#endif
 int right_alignment_SG(char * data1, int pos1, int len1,
		       char * data2, int pos2, int len2,
		       int XDROP,
		       /*out*/ int * p_gainscore,
		       /*out*/ int * p_last_pos1,
		       /*out*/ int * p_last_pos2,
		       Feature *feature
		       )
{
  int i,j;
  int max = 0;

  *p_gainscore = 0;
  *p_last_pos1 = pos1;
  *p_last_pos2 = pos2;

  M2(0,0) = 0;
  I2(0,0) = gp_cost_gap_opened -  gp_cost_gap_continued;
  
  for(i=1 ; i<len1+1 ; i++)
    M2(i,0) = -INFINITY_INT;
  
  for(i=1 ; i<len1+1 ; i++)
    I2(i,0) = I2(i-1,0) +  gp_cost_gap_continued;
  
  for(j=1 ; j<len2+1 ; j++){
    
    M2(0,j) =  -INFINITY_INT;
    I2(0,j) =  I2(0,j-1)  + gp_cost_gap_continued;
    max     =  M2(0,j);
    
    for(i=1 ; i<len1+1 ; i++){
      I2(i,j) = MAX (
		     MAX( M2(i,j-1) + gp_cost_gap_opened    , M2(i-1,j) + gp_cost_gap_opened )
		     ,
		     MAX( I2(i,j-1) + gp_cost_gap_continued ,  I2(i-1,j) + gp_cost_gap_continued )
		     );
      M2(i,j) = MAX ( M2(i-1,j-1) + S2(i,j), I2(i-1,j-1) + S2(i,j) ); 
      max     = MAX(max,M2(i,j));
      if (max > *p_gainscore) {
	*p_gainscore = max;
	*p_last_pos1 = pos1 + i;
	*p_last_pos2 = pos2 + j;
      }
    }
    if (max < -XDROP)
      return max;
  }
  return max;
}



#ifdef INLINE
 inline 
#endif
 int left_alignment_SG(char * data1, int pos1, int len1,
		       char * data2, int pos2, int len2,
		       int XDROP,
		       /*out*/ int * p_gainscore,
		       /*out*/ int * p_last_pos1,
		       /*out*/ int * p_last_pos2,
		       Feature *feature
		       )
{
  int i,j;
  int max = 0;

  *p_gainscore = 0;
  *p_last_pos1 = pos1;
  *p_last_pos2 = pos2;

  M2(0,0) = 0;
  I2(0,0) = gp_cost_gap_opened -  gp_cost_gap_continued;
  
  for(i=1 ; i<len1+1 ; i++)
    M2(i,0) = -INFINITY_INT;
  
  for(i=1 ; i<len1+1 ; i++)
    I2(i,0) = I2(i-1,0) +  gp_cost_gap_continued;
  
  for(j=1 ; j<len2+1 ; j++){
    
    M2(0,j) =  -INFINITY_INT;
    I2(0,j) =  I2(0,j-1)  + gp_cost_gap_continued;
    max     =  M2(0,j);
    
    for(i=1 ; i<len1+1 ; i++){
      I2(i,j) = MAX (
		     MAX( M2(i,j-1) + gp_cost_gap_opened    , M2(i-1,j) + gp_cost_gap_opened )
		     ,
		     MAX( I2(i,j-1) + gp_cost_gap_continued ,  I2(i-1,j) + gp_cost_gap_continued )
		     );
      M2(i,j) = MAX ( M2(i-1,j-1) + S2(-i+1,-j+1), I2(i-1,j-1) + S2(-i+1,-j+1) ); 
      max     = MAX(max,M2(i,j));
      if (max > *p_gainscore) {
	*p_gainscore = max;
	*p_last_pos1 = pos1 - i ;
	*p_last_pos2 = pos2 - j ;
      }
    }
    if (max < -XDROP)
      return max;
  }
  return max;
}



/*
 * left/rigth extension if score does not fall down under XDROP
 * 
 */

#ifdef INLINE
 inline 
#endif
 int left_alignment_SG_Diag (char * data1, int pos1, int len1,
			char * data2, int pos2, int len2, 
			int XDROP, 
			/*out*/ int * p_gainscore,
			/*out*/ int * p_last_pos1,
			/*out*/ int * p_last_pos2,
			Feature *feature)
{
  int i,j;
  int xdrop_current,xdrop_prev,xdrop_current2,xdrop_best=-INFINITY_INT;
  int d_i_prev,d_i;
  int m_i_prev,m_i;

  /* (1) initialisation */
  for(i=0;i<MIN(len1,len2);i++){
    D_I2(0,i) = -INFINITY_INT;/* FIXME */
    D_I2(1,i) = -INFINITY_INT;
    D_M2(0,i) = -INFINITY_INT;
    D_M2(1,i) = -INFINITY_INT;
  }
 
  D_M2(0,0) = 0;   
  *(p_gainscore) = 0;
  *(p_last_pos1) = pos1;
  *(p_last_pos2) = pos2;
  
  xdrop_current2 = 0;
  xdrop_current  = -INFINITY_INT;
  xdrop_best     = 0;
  
  /* (2) compute (stop if Xdrop) */
  for(j=0;j<MIN(len1,len2);j++) {
    
    d_i = -INFINITY_INT;
    m_i = -INFINITY_INT;
    
    xdrop_prev    = xdrop_current;
    xdrop_current = -INFINITY_INT; 
    
    for(i=0;i<=j+1;i++){
       
      d_i_prev = d_i;
      m_i_prev = m_i;
      
      if (i<j){
	d_i      = D_I2(j+1,i);
	m_i      = D_M2(j+1,i);
      }else{
	d_i      = -INFINITY_INT;
	m_i      = -INFINITY_INT;
      }

      /* compute  d_i(j+1) */
      D_I2(j+1,i) = MAX(
			MAX( ((i) ? D_M2(j,i-1) : -INFINITY_INT ), D_M2(j,i)) + gp_cost_gap_opened
			,
			MAX( ((i) ? D_I2(j,i-1) : -INFINITY_INT ), D_I2(j,i)) + gp_cost_gap_continued
		       );

      /*  compute  m_i(j+1) */
      D_M2(j+1,i) = MAX ( 
			 m_i_prev  + D_S2(-i,(-j+i-1)), 
			 d_i_prev  + D_S2(-i,(-j+i-1))
			 );
      
      
      xdrop_current = MAX(xdrop_current,D_M2(j+1,i)); 
      
    }/* i */
    if (j&1)
      xdrop_current2 = MAX(xdrop_current,xdrop_prev);
    
    /* printf("XDROP %i\n",xdrop_current2); */
    if (xdrop_current2 <= -XDROP){
      /* set the parameters beeing returned */
      return xdrop_current2;
    }else{
      if(xdrop_current > xdrop_best){
	xdrop_best     = xdrop_current;
	*(p_gainscore) = xdrop_best;
	for(i=0;i<=j+1;i++)
	  if (D_M2(j+1,i) == xdrop_best){
	    *(p_last_pos1) = pos1 - i;
	    *(p_last_pos2) = pos2 - j + i  - 1 ;
	    break;
	  }
      }
    }  
  }
  return 0;
}




#ifdef INLINE
 inline 
#endif
 int right_alignment_SG_Diag(char * data1, int pos1, int len1,
			char * data2, int pos2, int len2, 
			int XDROP, 
			/*out*/ int * p_gainscore,
			/*out*/ int * p_last_pos1,
			/*out*/ int * p_last_pos2,
			Feature *feature)
{
  int i,j;
  int xdrop_current,xdrop_prev,xdrop_current2,xdrop_best=-INFINITY_INT;
  int d_i_prev,d_i;
  int m_i_prev,m_i;
  

  /* (1) initialisation */
  for(i=0;i<MIN(len1,len2);i++){
    D_I2(0,i) = -INFINITY_INT;
    D_I2(1,i) = -INFINITY_INT;
    D_M2(0,i) = -INFINITY_INT;
    D_M2(1,i) = -INFINITY_INT;
  }
 
  D_M2(0,0) = 0;   
  *(p_gainscore) = 0;
  *(p_last_pos1) = pos1;
  *(p_last_pos2) = pos2;
  
  xdrop_current2 = 0;
  xdrop_current  = -INFINITY_INT;
  xdrop_best     = 0;
  
  /* (2) compute (stop if Xdrop) */
  for(j=0;j<MIN(len1,len2);j++){/* for each diagonal */
    
    d_i = -INFINITY_INT;
    m_i = -INFINITY_INT;
    
    xdrop_prev    = xdrop_current;
    xdrop_current = -INFINITY_INT;
    
    for(i=0;i<=j+1;i++){
       
      d_i_prev = d_i;
      m_i_prev = m_i;
      
      if (i<j){
	d_i      = D_I2(j+1/*j-1*/,i);
	m_i      = D_M2(j+1/*j-1*/,i);
      }else{
	d_i      = -INFINITY_INT;
	m_i      = -INFINITY_INT;
      }

      printf("j = %i, i = %i :  \t m_i = %i:%i \t d_i = %i:%i \n",j,i,m_i,m_i_prev,d_i,d_i_prev);
      
      
      /* compute d_i(j+1) */
      D_I2(j+1,i) = MAX(
			MAX( ((i) ? D_M2(j,i-1) : -INFINITY_INT ), D_M2(j,i)) + gp_cost_gap_opened
			,
			MAX( ((i) ? D_I2(j,i-1) : -INFINITY_INT ), D_I2(j,i)) + gp_cost_gap_continued
		       );

      /* compute m_i(j+1) */
      D_M2(j+1,i) = MAX ( 
			 m_i_prev  + D_S2(i,(j-i+1)), 
			 d_i_prev  + D_S2(i,(j-i+1))
			 );
      
      
      xdrop_current = MAX(xdrop_current,D_M2(j+1,i));
    }/* i */
    if (j&1)
	xdrop_current2 = MAX(xdrop_current,xdrop_prev);
    if (xdrop_current2 <= -XDROP){
      /* mise en place parametres retour */
      return xdrop_current2;
    }else{
      if(xdrop_current > xdrop_best){
	xdrop_best = xdrop_current;
	*(p_gainscore) = xdrop_best;
	for(i=0;i<=j+1;i++)
	  if (D_M2(j+1,i) == xdrop_best){
	    *(p_last_pos1) = pos1 + i  +  1;
	    *(p_last_pos2) = pos2 + j - i + 2;
	    break;
	  }
      }
    }  
  }
  return 0;
}

#ifdef INLINE
 inline 
#endif	
int gp_test=0;
 int left_alignment_SG_Lz(char * data1, int pos1, int len1,
			  char * data2, int pos2, int len2, 
			  int XDROP, 
			  /*out*/ int * p_gainscore,
			  /*out*/ int * p_last_pos1,
			  /*out*/ int * p_last_pos2,
			  Feature *feature){
   int len = MIN(len1,len2) - gp_delta_stat;
   int last_indels = 0, indels = 0;
   int last_i      = 0, i      = 0;
   int gainscore = 0;
  

   while(i<len){
     if (indels!=0)
       gainscore = *p_gainscore +  gp_cost_gap_opened + ABS(indels - last_indels) * gp_cost_gap_continued;

     for(i = last_i+1; i<len; i++ ) {
       gainscore += D_S2(-i,-i-indels);

       if (gainscore < -XDROP) 
	break;
      
       /* single file stop */
       if (data1 + pos1 - i == data2 + pos2 - i - indels)
	 break;
       
       if (gainscore > *p_gainscore) {
	 
	 *p_gainscore  = gainscore;
	 *p_last_pos1  = pos1-i;
	 *p_last_pos2  = pos2-i+indels;
	 last_i        = i;
	 last_indels   = indels;
       }
#ifdef DEBUG_ALIGNMENT_EXTENSION
       fprintf(stderr,"pos1-i=%d pos2-i+indels=%d gainscore=%d  [p_last_pos1=%d p_last_pos1=%d p_gainscore=%d] (%c==%c)\n",
	       pos1-i,pos2-i+indels,gainscore,
	       *p_last_pos1,*p_last_pos2, *p_gainscore,
	       lookup[data1[pos1-i]],lookup[data2[pos2-i]]);
#endif
     }/* for i */
    
     if (indels > 0){
       indels = -indels;
     } else {
       indels = -indels+1;
     }
     if (indels > gp_delta_stat) 
       break;
     
   }/* while */

   return 1;
 }





#ifdef INLINE
 inline 
#endif
 int right_alignment_SG_Lz(char * data1, int pos1, int len1,
			   char * data2, int pos2, int len2, 
			   int XDROP, 
			   /*out*/ int * p_gainscore,
			   /*out*/ int * p_last_pos1,
			   /*out*/ int * p_last_pos2,
			   Feature *feature
			   )
{
  int len = MIN(len1,len2) - gp_delta_stat;
  int last_indels = 0, indels = 0;
  int last_i      = 0, i      = 0; 
  int gainscore = 0;

  while(i<len){
    if (indels!=0)
      gainscore = *p_gainscore +  gp_cost_gap_opened + ABS(indels - last_indels) * gp_cost_gap_continued;

    for(i = last_i; i<len; i++ ) {
     
      gainscore += D_S2(i,i+indels);
      /* fprintf(stderr,"%d  %c %c\n",gainscore,lookup[*(data1+pos1+i)],lookup[*(data2+pos2+i)]); */
      if (gainscore < -XDROP) 
	break;
      
      /* single file stop */
      if (data1 + pos1 + i == data2 + pos2 + i + indels)
	break;
      
      if (gainscore > *p_gainscore) {
	  
	*p_last_pos1  = pos1+i+1;
	*p_last_pos2  = pos2+i+indels+1;
	last_i        = i;
	*p_gainscore  = gainscore;
	last_indels   = indels;
      }
    }/* for i */

    if (indels > 0){
      indels = -indels;
    } else {
      indels = -indels+1;
    }
    if (indels > gp_delta_stat)
      break;
  }/* while */

  return 1;
}





int right_alignment_SG_Border(char * data1, int pos1, int len1, 
			      char * data2, int pos2, int len2,
			      int XDROP, 
			      /*out*/ int * p_gainscore,
			      /*out*/ int * p_last_pos1,
			      /*out*/ int * p_last_pos2,
			      Feature *feature)
{
  /* band alignment length anw width*/
  int bandwidth  = gp_delta_stat;
 
  /* various switch elements to access information */
  int row1       = 2*bandwidth;
  int row2       = 2*row1;
  
  /* dp tables */
  int * bm0 = (int *)feature->buffer01;
  int * bm1 = (int *)feature->buffer01+row1;
  int * bi0 = (int *)feature->buffer01+row2;
  int * bi1 = (int *)feature->buffer01+row2+row1;
  
  /* maximal and minimal positions taken into account */
  int j,k;
  int jmin = 0, jmax = row1;
  
  for(j=0;j<row1;j++)
    bm0[j] = bi0[j] = (gp_cost_gap_opened) + ABS(j-bandwidth)*gp_cost_gap_continued;
  
  bm0[bandwidth] = bi0[bandwidth] = 0;



  if (len1 < len2){
    
    for(k=1 ; k< len1; k++){ 
     
      if(k&1){	
	/* M(j) full vectorisable loop */
	for(j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data1 + pos1)[k])] [(int)(data2 + pos2)[j+k-bandwidth]];
	  bm1[j] = MAX( bm0[j] , bi0[j] ) + d; 
	  
	  if (bm1[j] < *p_gainscore) {
	    *p_gainscore = bm1[j];
	    *p_last_pos1 = pos1 + k;
	    *p_last_pos2 = pos2 + j + k - bandwidth;
	  }
	  
	}	
	
	/* MinxMax evaluations */
	j=jmin;	
	while( bm1[j++] < -XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm1[j--] < -XDROP && j>0)
	  jmax--;
	if (jmax<jmin)
	  return -XDROP-1;
  

	/* I(j) in two half loops */
	bi1[bandwidth] = MAX(MAX(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , MAX(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi1[j-1] = MAX(
			 MAX(bi0[j-1],bi1[j]) + gp_cost_gap_continued,
			 MAX(bm0[j-1],bm1[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi1[j] = MAX(
		       MAX(bi0[j],bi1[j-1]) + gp_cost_gap_continued,
		       MAX(bm0[j],bm1[j-1]) + gp_cost_gap_opened
		       );
	}

      }else{/* k&1*/
          
      	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data1 + pos1)[k])][(int)(data2 + pos2)[j+k-bandwidth]];
	  bm0[j] = MAX( bm1[j] , bi1[j] ) + d;

	  if (bm0[j] < *p_gainscore) {
	    *p_gainscore = bm0[j];
	    *p_last_pos1 = pos1 + k;
	    *p_last_pos2 = pos2 + j + k - bandwidth;
	  }

	}

	/* MinxMax evaluations */
	j=jmin;	
	while( bm0[j++] < -XDROP && j<jmax)
	  jmin++;
	j=jmax;	
	while( bm0[j--] < -XDROP && j>0)
	  jmax--;
	if (jmax<jmin)
	  return -XDROP-1;
	

	/* I(j) in two half loops */
	bi0[bandwidth] = MAX(MAX(bi1[bandwidth-1],bi1[bandwidth+1]) + gp_cost_gap_continued , MAX(bm1[bandwidth-1],bm1[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi0[j-1] = MAX(
			 MAX(bi1[j-1],bi0[j]) + gp_cost_gap_continued,
			 MAX(bm1[j-1],bm0[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi0[j] = MAX(
		       MAX(bi1[j],bi0[j-1]) + gp_cost_gap_continued,
		       MAX(bm1[j],bm0[j-1]) + gp_cost_gap_opened
		       ); 
	}
      }/* k&1*/
    }/* k */

  }else{
    
    for( k=1 ; k< len2; k++){ 
          
      if(k&1){	
	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[((int)(data1 + pos1)[j+k-bandwidth])][(int)(data2 + pos2)[k]];
	  bm1[j] = MAX( bm0[j] , bi0[j] ) + d;

	  if (bm1[j] < *p_gainscore) {
	    *p_gainscore = bm1[j];
	    *p_last_pos1 = pos1 + k;
	    *p_last_pos2 = pos2 + j + k - bandwidth;
	  }

	}	

	/* MinxMax evaluations */
	 j=jmin;	
	while( bm1[j++] < -XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm1[j--] < -XDROP && j>0)
	  jmax--;
	
	if (jmax<jmin)
	  return -XDROP-1;

	/* I(j) in two half loops (each on is vectorisable on "non stupid" compilers) */
	bi1[bandwidth] = MAX(MAX(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , MAX(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi1[j-1] = MAX(
			 MAX(bi0[j-1],bi1[j]) + gp_cost_gap_continued,
			 MAX(bm0[j-1],bm1[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi1[j] = MAX(
		       MAX(bi0[j],bi1[j-1]) + gp_cost_gap_continued,
		       MAX(bm0[j],bm1[j-1]) + gp_cost_gap_opened
		       );
	}

      }else{/* k&1*/
          
      	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data2 + pos2)[k])<<2][(int)(data1 + pos1)[j+k-bandwidth]];
	  bm0[j] =  MAX( bm1[j] , bi1[j] ) + d; 

	  if (bm0[j] < *p_gainscore) {
	    *p_gainscore = bm0[j];
	    *p_last_pos1 = pos1 + k;
	    *p_last_pos2 = pos2 + j + k - bandwidth;
	  }

	}	

	/* MinxMax evaluations */
	j=jmin;
	while( bm0[j++] < -XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm0[j--] < -XDROP && j>0)
	  jmax--;
	
	if (jmax<jmin)
	  return -XDROP-1;
	
	/* I(j) in two half loops (each on is vectorisable on "non stupid" compilers) */
	bi0[bandwidth] = MAX(MAX(bi1[bandwidth-1],bi1[bandwidth+1]) + gp_cost_gap_continued , MAX(bm1[bandwidth-1],bm1[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi0[j-1] = MAX(
			 MAX(bi1[j-1],bi0[j]) + gp_cost_gap_continued,
			 MAX(bm1[j-1],bm0[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi0[j] = MAX(
		       MAX(bi1[j],bi0[j-1]) + gp_cost_gap_continued,
		       MAX(bm1[j],bm0[j-1]) + gp_cost_gap_opened
		       ); 
	}
      }/* k&1*/
    }/* k */
 
  }
  return (MAX(len1,len2)&1)?(bm0[bandwidth]):(bm1[bandwidth]);
}





int left_alignment_SG_Border(char * data1, int pos1, int len1, 
			     char * data2, int pos2, int len2,
			     int XDROP, 
			      /*out*/ int * p_gainscore,
			      /*out*/ int * p_last_pos1,
			      /*out*/ int * p_last_pos2,
			     Feature *feature)
{
  /* band alignment length anw width*/
  int bandwidth  = gp_delta_stat;
  
 
  /* various switch elements to access information */
  int row1       = 2*bandwidth;
  int row2       = 2*row1;
  
  /* dp tables */
  int * bm0 = (int *)feature->buffer01;
  int * bm1 = (int *)feature->buffer01+row1;
  int * bi0 = (int *)feature->buffer01+row2;
  int * bi1 = (int *)feature->buffer01+row2+row1;
  
  /* maximal and minimal positions taken into account */
  int j,k;
  int jmin = 0, jmax = row1;
  
  for(j=0;j<row1;j++)
    bm0[j] = bi0[j] = (gp_cost_gap_opened) + ABS(j-bandwidth)*gp_cost_gap_continued;
  
  bm0[bandwidth] = bi0[bandwidth] = 0;

  if (len1 < len2){
    
    for(k=1 ; k< len1; k++){ 
     
      if(k&1){	
	/* M(j) full vectorisable loop */
	for(j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data1 + pos1)[-k])][(int)(data2 + pos2)[j-k-bandwidth]];
	  bm1[j] = MAX( bm0[j] , bi0[j] ) + d; 
	  
	  if (bm1[j] > *p_gainscore) {
	    *p_gainscore = bm1[j];
	    *p_last_pos1 = pos1 - k;
	    *p_last_pos2 = pos2 + j - k - bandwidth;
	  }
	  
	}	
	
	/* MinxMax evaluations */
	j=jmin;	
	while( bm1[j++] < -XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm1[j--] < -XDROP && j>0)
	  jmax--;
	if (jmax<jmin)
	  return -XDROP-1;
  

	/* I(j) in two half loops */
	bi1[bandwidth] = MAX(
			     MAX(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , 
			     MAX(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened
			     );	
	for( j=bandwidth ; j>jmin ; j--){
	  bi1[j-1] = MAX(
			 MAX(bi0[j-1],bi1[j]) + gp_cost_gap_continued,
			 MAX(bm0[j-1],bm1[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi1[j] = MAX(
		       MAX(bi0[j],bi1[j-1]) + gp_cost_gap_continued,
		       MAX(bm0[j],bm1[j-1]) + gp_cost_gap_opened
		       );
	}

      }else{/* k&1*/
          
      	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data1 + pos1)[-k])][(int)(data2 + pos2)[j-k-bandwidth]];
	  bm0[j] = MAX( bm1[j] , bi1[j] ) + d;

	  if (bm0[j] > *p_gainscore) {
	    *p_gainscore = bm0[j];
	    *p_last_pos1 = pos1 - k;
	    *p_last_pos2 = pos2 + j - k - bandwidth;
	  }

	}

	/* MinxMax evaluations */
	j=jmin;	
	while( bm0[j++] < -XDROP && j<jmax)
	  jmin++;
	j=jmax;	
	while( bm0[j--] < -XDROP && j>0)
	  jmax--;
	if (jmax<jmin)
	  return -XDROP-1;
	

	/* I(j) in two half loops */
	bi0[bandwidth] = MAX(
			     MAX(bi1[bandwidth-1],bi1[bandwidth+1]) + gp_cost_gap_continued , 
			     MAX(bm1[bandwidth-1],bm1[bandwidth+1]) + gp_cost_gap_opened
			     );	
	for( j=bandwidth ; j>jmin ; j--){
	  bi0[j-1] = MAX(
			 MAX(bi1[j-1],bi0[j]) + gp_cost_gap_continued,
			 MAX(bm1[j-1],bm0[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi0[j] = MAX(
		       MAX(bi1[j],bi0[j-1]) + gp_cost_gap_continued,
		       MAX(bm1[j],bm0[j-1]) + gp_cost_gap_opened
		       ); 
	}
      }/* k&1*/
    }/* k */

  }else{
    
    for( k = 1 ; k < len2; k++){ 
      
      if(k&1){	
	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data2 + pos2)[-k])][(int)(data1 + pos1)[j-k-bandwidth]];
	  bm1[j] = MAX( bm0[j] , bi0[j] ) + d;

	  if (bm1[j] > *p_gainscore) {
	    *p_gainscore = bm1[j];
	    *p_last_pos1 = pos1 - k;
	    *p_last_pos2 = pos2 + j - k - bandwidth;
	  }

	}	

	/* MinxMax evaluations */
	 j=jmin;	
	while( bm1[j++] < -XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm1[j--] < -XDROP && j>0)
	  jmax--;
	
	if (jmax<jmin)
	  return -XDROP-1;

	/* I(j) in two half loops (each on is vectorisable on "non stupid" compilers) */
	bi1[bandwidth] = MAX(MAX(bi0[bandwidth-1],bi0[bandwidth+1]) + gp_cost_gap_continued , MAX(bm0[bandwidth-1],bm0[bandwidth+1]) + gp_cost_gap_opened);
	for( j=bandwidth ; j>jmin ; j--){
	  bi1[j-1] = MAX(
			 MAX(bi0[j-1],bi1[j]) + gp_cost_gap_continued,
			 MAX(bm0[j-1],bm1[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi1[j] = MAX(
		       MAX(bi0[j],bi1[j-1]) + gp_cost_gap_continued,
		       MAX(bm0[j],bm1[j-1]) + gp_cost_gap_opened
		       );
	}

      }else{/* k&1*/
          
      	/* M(j) full vectorisable loop */
	for( j=jmin ; j<jmax ; j++ ){	  
	  int d = gp_substitution_matrix[(int)((data2 + pos2)[-k])][(int)(data1 + pos1)[j-k-bandwidth]];
	  bm0[j] =  MAX( bm1[j] , bi1[j] ) + d; 

	  if (bm0[j] > *p_gainscore) {
	    *p_gainscore = bm0[j];
	    *p_last_pos1 = pos1 - k;
	    *p_last_pos2 = pos2 + j - k - bandwidth;
	  }

	}	

	/* MinxMax evaluations */
	j=jmin;
	while( bm0[j++] < -XDROP && j<jmax)
	  jmin++;
	j=jmax;
	while( bm0[j--] < -XDROP && j>0)
	  jmax--;
	
	if (jmax<jmin)
	  return -XDROP-1;
	
	/* I(j) in two half loops (each on is vectorisable on "non stupid" compilers) */
	bi0[bandwidth] = MAX(MAX(bi1[bandwidth-1],bi1[bandwidth+1]) + gp_cost_gap_continued , MAX(bm1[bandwidth-1],bm1[bandwidth+1]) + gp_cost_gap_opened);	
	for( j=bandwidth ; j>jmin ; j--){
	  bi0[j-1] = MAX(
			 MAX(bi1[j-1],bi0[j]) + gp_cost_gap_continued,
			 MAX(bm1[j-1],bm0[j]) + gp_cost_gap_opened
			 );
	}
	for( j=bandwidth; j<jmax ; j++){
	  bi0[j] = MAX(
		       MAX(bi1[j],bi0[j-1]) + gp_cost_gap_continued,
		       MAX(bm1[j],bm0[j-1]) + gp_cost_gap_opened
		       ); 
	}
      }/* k&1*/
    }/* k */
 
  }
  return (MAX(len1,len2)&1)?(bm0[bandwidth]):(bm1[bandwidth]);
}





/*
 *  [II]
 *
 *   "slow" alignment functions used to compute statistical
 *   alignements parameters.
 *
 ***********************************************************************/




#define M3(i,j) (mt[(i)+ (j)*(len1+1)])
#define I3(i,j) (in[(i)+ (j)*(len1+1)])
#define S3(i,j) S2(i,j)
#define D3(k)   (bk[(k)])

/*
 * 1) complexit linguistique : nombre de mots de taille 1,2, ... , n
 * 2) entropy : - sum p_i log (1/p_i) pour trinucleotides : 
 *    tableau de 4^s compteurs, un entier pour le shift-or d'acces
 *    
 *
 *
 *
 */


typedef enum backtrack_t { _B_EMPTY_,
			   STEP_LEFT,
			   STEP_RIGHT,
			   DIAG_EGAL,
			   DIAG_DIFF } backtrack;

typedef enum backtrack_SG_t { _BSG_EMPTY_,
			      M_TO_M_SUBS,
			      I_TO_M_SUBS, 
			      FIRST_DEL_LEFT,
			      FIRST_DEL_RIGHT,
			      CONT_DEL_LEFT,
			      CONT_DEL_RIGHT } backtrack_SG;


#define WORDCOUNTANDSTAT(count,letter,word) {(word) <<= 2; (word) |= backcode[(int)(letter)]; (word) &= 0xfc00003f;  if (((word) & 0x3f) == (word)) count[(word)]++;}

#define PAIROFWORDCOUNANDSTAT(count,word1,word2) { if ((word1) > 0 && (word1) < 64 && (word2) > 0 && (word2) < 64) ((count)[word1][word2])++;}


int alignment_SG_stats(char * data1, int pos1, int len1,
		       char * data2, int pos2, int len2,
		       /*out*/int * transindels      /*[3]*/,
		       /*out*/int * trinomial_count1 /*[3]*/,
		       /*out*/int * trinomial_count2 /*[3]*/,
		       /*out*/int * mutatedword1,
		       /*out*/int * mutatedword2,
		       /*out*/int * nonmutatedword) 
{
  
  int i,j,k;
  int i1,i2;
  int result;
  int *mt, *in;
  backtrack_SG * bk;
    
  mt = (int*) MALLOC( (unsigned) (len1+1) * (len2+1) * sizeof(int));
  ASSERT(mt,alignment_SG_stats);
  in = (int*) MALLOC( (unsigned) (len1+1) * (len2+1) * sizeof(int));  
  ASSERT(in,alignment_SG_stats);
  bk = (backtrack_SG*) MALLOC( (unsigned) (len1+len2+1) * sizeof(backtrack_SG));
  ASSERT(bk,alignment_SG_stats);
  
  /* [1] compute */

  M3(0,0) = 0;
  
  I3(0,0) = gp_cost_gap_opened - gp_cost_gap_continued;
  
  for(i=1;i<len1+1; i++)
    M3(i,0) = -INFINITY_INT;
  
  for(i=1 ; i<len1+1 ; i++)
    I3(i,0) = I3(i-1,0) + gp_cost_gap_continued;
  

  for(j=1 ; j<len2+1 ; j++){
  
    M3(0,j) =  -INFINITY_INT;
    I3(0,j) =  gp_cost_gap_opened  + (j-1)*gp_cost_gap_continued;
    
    for(i=1;i<len1+1;i++){
      I3(i,j) = MAX (
		     MAX( M3(i,j-1) + gp_cost_gap_opened    , M3(i-1,j) + gp_cost_gap_opened )
		     ,
		     MAX( I3(i,j-1) + gp_cost_gap_continued , I3(i-1,j) + gp_cost_gap_continued )
		     );
      M3(i,j) = MAX ( M3(i-1,j-1) + S3(i,j), I3(i-1,j-1) + S3(i,j) ); 
    }
  }
  
  result = MAX( M3(len1,len2) , I3(len1,len2) );
 
  /* [2] backtrace */
 
  i = len1;
  j = len2;
  k = 0;
  
  if (M3(len1,len2) < I3(len1,len2))
    D3(k++) = I_TO_M_SUBS;
  else
    D3(k++) = M_TO_M_SUBS;
  
  
  while (i > 0 && j > 0){
    
    if ( D3(k-1) == M_TO_M_SUBS || D3(k-1) == FIRST_DEL_LEFT || D3(k-1) == FIRST_DEL_RIGHT ){
      
      /* no gap to open  */
      if ( M3(i,j) == M3(i-1,j-1) + S3(i,j) ){ 
	i--;
	j--;
	D3(k++) = M_TO_M_SUBS;
      }else{	
	if ( M3(i,j) == I3(i-1,j-1) + S3(i,j) ){ 
	  i--;
	  j--;
	  D3(k++) = I_TO_M_SUBS;
	}else{
	  _WARNING("alignment_SG_stats() 1");
	}
      } 
    }else{   
      /* need to open a gap */
      if ( I3(i,j) == M3(i,j-1) + gp_cost_gap_opened ){
	j--;
	D3(k++) =  FIRST_DEL_RIGHT ;
      }else{
	if ( I3(i,j) == M3(i-1,j) + gp_cost_gap_opened ){
	  i--;
	  D3(k++) =  FIRST_DEL_LEFT ;
	}else{
	  if ( I3(i,j) == I3(i,j-1) + gp_cost_gap_continued ){
	    j--;
	    D3(k++) =  CONT_DEL_RIGHT ;
	  }else{
	    if ( I3(i,j) == I3(i-1,j) + gp_cost_gap_continued ){
	      i--;
	      D3(k++) =  CONT_DEL_LEFT ;
	    }else{
	      _WARNING("alignment_SG_stats() 2");
	    }
	  } 
	}
      }
    }
  }/* while */

  
  if (j > 0) {
    while(j > 1) {
      j--;
      D3(k++) = CONT_DEL_RIGHT;
    }  
    D3(k++) = FIRST_DEL_RIGHT;
  }
  
  if (i > 0) {
    while(i > 1) {
      i--;
      D3(k++) = CONT_DEL_LEFT;
    }
    D3(k++) = FIRST_DEL_LEFT;
  }   

  FREE(mt,(len1+1) * (len2+1) * sizeof(int));
  FREE(in,(len1+1) * (len2+1) * sizeof(int));


  /* [3] backtrace to compute some statistics */

  i1 = 0; i2 = 0; k--; 
  
  while (i1 < len1 || i2 < len2 ){

    switch(D3(k--)){
    case M_TO_M_SUBS:
    case I_TO_M_SUBS:
      
      if ( TNORM(data1[pos1+i1]) != TNORM(data2[pos2+i2]) ) { /* substitution */
	(trinomial_count1[(pos1+i1)%3])++;
	(trinomial_count2[(pos2+i2)%3])++;
	*nonmutatedword = 0xfc000000;
	if ((TNORM(data1[pos1+i1])^TNORM(data2[pos2+i2])) == 2)
	  (transindels[0])++; /* transition */
	else
	  (transindels[1])++; /* transversion */
      } else {
	WORDCOUNTANDSTAT(gv_nb_non_mutated_triplet_count,data1[pos1+i1],*nonmutatedword);
      }
      WORDCOUNTANDSTAT(gv_nb_triplet_count1,data1[pos1+i1],*mutatedword1);
      WORDCOUNTANDSTAT(gv_nb_triplet_count2,data2[pos2+i2],*mutatedword2);
      PAIROFWORDCOUNANDSTAT(gv_nb_pair_of_triplets,*mutatedword1,*mutatedword2);
      
      i1++;
      i2++;
      break;
      
    case FIRST_DEL_LEFT:
    case CONT_DEL_LEFT:
      *nonmutatedword = 0xfc000000;
      transindels[2]++;
      i1++;
      break;
      
    case FIRST_DEL_RIGHT:
    case CONT_DEL_RIGHT:
      *nonmutatedword = 0xfc000000;
      transindels[2]++;
      i2++;
      break;
      
    default:
      _WARNING("alignment_SG_stats() 3");

    }
    
  }/* while */
  
  FREE(bk, (len1+len2+1) * sizeof(backtrack_SG));
  return result; 
}




int alignment_SG_stats_Strait(char * data1,  int pos1, char * data2,  int pos2, int len,
			      /*out*/int * transindels      /*[3]*/,
			      /*out*/int * trinomial_count1 /*[3]*/,
			      /*out*/int * trinomial_count2 /*[3]*/,
			      /*out*/int * mutatedword1,
			      /*out*/int * mutatedword2,
			      /*out*/int * nonmutatedword){

  int i;
  for ( i=0; i<len; i++ ) { 
    if ( TNORM(data1[pos1+i]) != TNORM(data2[pos2+i]) ) { /* substitution */
	(trinomial_count1[(pos1+i)%3])++;
	(trinomial_count2[(pos2+i)%3])++;
	*nonmutatedword = 0xfc000000;
	if (((data1[pos1+i])^(data2[pos2+i])) == 2)
	  (transindels[0])++; /* transition */
	else
	  (transindels[1])++; /* transverstion */
    } else {
      WORDCOUNTANDSTAT(gv_nb_non_mutated_triplet_count,data1[pos1+i],*nonmutatedword);
    }
    WORDCOUNTANDSTAT(gv_nb_triplet_count1,data1[pos1+i],*mutatedword1);
    WORDCOUNTANDSTAT(gv_nb_triplet_count2,data2[pos2+i],*mutatedword2);
    PAIROFWORDCOUNANDSTAT(gv_nb_pair_of_triplets,*mutatedword1,*mutatedword2);
  }
  return 0;
}




int alignment_SG_stats_on_MA(char * data1, char * data2,
			     MA * ma,
			     int left_correction,
			     /*out*/int * transindels      /*[3]*/,
			     /*out*/int * trinomial_count1 /*[3]*/,
			     /*out*/int * trinomial_count2 /*[3]*/) 
{
  

  int left_size,right_size;
  tuple * t = ma->first_tuple, * t_prev = ma->first_tuple ;
  int mutatedword1    = 0xfc000000;
  int mutatedword2    = 0xfc000000;
  int nonmutatedword  = 0xfc000000;

  /* [1] Compute alignment before first tuple */
  
  left_size  = TBL_POS(t) - ma->left_pos_begin  ;
  right_size = TBR_POS(t) - ma->right_pos_begin ;
  
  if ( left_size > 0 && right_size > 0)
    alignment_SG_stats(data1, ma->left_pos_begin , left_size ,
		       data2, ma->right_pos_begin, right_size,
		       transindels,trinomial_count1,trinomial_count2,
		       &mutatedword1,&mutatedword2,&nonmutatedword);
  
#ifdef DEBUG_STATS
  fprintf(stderr,"Stat MA=(%d,%d)(%d,%d) -> extension gauche triplet=(%d,%d,%d) tr=%d tv=%d state(%d,%d,%d)\n",ma->left_pos_begin,
	  ma->left_pos_end,ma->right_pos_begin,ma->right_pos_end,
	  trinomial_count1[0],trinomial_count1[1],trinomial_count1[2],
	  transindels[0],transindels[1],
	  trinomial_count2[0],trinomial_count2[1],trinomial_count2[2],
	  );
#endif 

  /* [2] Compute alignment for several tuples */
 
  while(t != NULL){
    /* (1) compute inside a tuple */
    alignment_SG_stats_Strait(data1,TBL_POS(t),
			      data2,TBR_POS(t),TSIZE(t),
			      transindels,trinomial_count1,trinomial_count2,
			      &mutatedword1,&mutatedword2,&nonmutatedword);
    
#ifdef DEBUG_STATS
    fprintf(stderr,"Stat MA=(%d,%d)(%d,%d) -> tuple triplet=(%d,%d,%d) tr=%d tv=%d state(%d,%d,%d)\n",ma->left_pos_begin,
	    ma->left_pos_end,ma->right_pos_begin,ma->right_pos_end,
	    trinomial_count1[0],trinomial_count1[1],trinomial_count1[2],
	    transindels[0],transindels[1],
	    trinomial_count2[0],trinomial_count2[1],trinomial_count2[2]
	    );
#endif
    if (t->next){
      /* (2) alignment between (t) et (t->next) */
      alignment_SG_stats(data1,TEL_POS(t),TGAP_L(t,t->next),
			 data2,TER_POS(t),TGAP_R(t,t->next),
			 transindels,trinomial_count1,trinomial_count2,
			 &mutatedword1,&mutatedword2,&nonmutatedword);
      

#ifdef DEBUG_STATS
    fprintf(stderr,"Stat MA=(%d,%d)(%d,%d) -> between triplet=(%d,%d,%d) tr=%d tv=%d state(%d,%d,%d)\n",ma->left_pos_begin,
	    ma->left_pos_end,ma->right_pos_begin,ma->right_pos_end,
	    trinomial_count1[0],trinomial_count1[1],trinomial_count1[2],
	    transindels[0],transindels[1],
	    trinomial_count2[0],trinomial_count2[1],trinomial_count2[2],
	    &mutatedword1,&mutatedword2,&nonmutatedword);
#endif
    }/* if */
    t_prev = t;
    t = t->next;
  }/* while */

  /* [3] Compute alignement after last tuple */
  
  left_size  =  ma->left_pos_end  - TEL_POS(t_prev) ;
  right_size =  ma->right_pos_end - TER_POS(t_prev) ;
    
  if ( left_size > 0 && right_size> 0)
    alignment_SG_stats(data1, TEL_POS(t_prev), left_size ,
		       data2, TER_POS(t_prev), right_size,
		       transindels,trinomial_count1,trinomial_count2,
		       &mutatedword1,&mutatedword2,&nonmutatedword);
#ifdef DEBUG_STATS
    fprintf(stderr,"Stat MA=(%d,%d)(%d,%d) -> end MA =(%d,%d,%d) tr=%d tv=%d\n",ma->left_pos_begin,
	    ma->left_pos_end,ma->right_pos_begin,ma->right_pos_end,
	    trinomial_count1[0],trinomial_count1[1],trinomial_count1[2],
	    transindels[0],transindels[1]);
#endif
    return 0;
 }




int alignment_SG_stats_on_strait_MA(char * data1, char * data2,
 			     MA * ma,
			     int left_correction,
			     /*out*/int * transindels      /*[3]*/,
			     /*out*/int * trinomial_count1 /*[3]*/,
                             /*out*/int * trinomial_count2 /*[3]*/) 
{
  
  

   tuple * t = ma->first_tuple;
   int mutatedword1    = 0xfc000000;
   int mutatedword2    = 0xfc000000;
   int nonmutatedword  = 0xfc000000;

   /* [1] Compute alignment for several tuples */
   while(t != NULL){
     /* compute inside a tuple */
     alignment_SG_stats_Strait(data1,TBL_POS(t),
	 		       data2,TBR_POS(t),TSIZE(t),
			       transindels,trinomial_count1,trinomial_count2,
                               &mutatedword1,&mutatedword2,&nonmutatedword);

			       
#ifdef DEBUG_STATS
    fprintf(stderr,"Stat MA=(%d,%d)(%d,%d) -> tuple triplet=(%d,%d,%d) tr=%d tv=%d state(%d,%d,%d)\n",ma->left_pos_begin,
	    ma->left_pos_end,ma->right_pos_begin,ma->right_pos_end,
	    trinomial_count1[0],trinomial_count1[1],trinomial_count1[2],
	    transindels[0],transindels[1],
	    trinomial_count2[0],trinomial_count2[1],trinomial_count2[2]
	    );
#endif

     t = t->next;
   }/* while */
   return 0;
}




/*
 *  [III]
 *
 *   "slow"  alignment methods used to visualise
 *   provided alignments 
 *
 ***********************************************************************/



/*
 *
 *  display buffer composed of 5 line with NBL chars
 *
 */


int line_pos = 0;
char line_dpos1[NBL+1];
char line_d1[NBL+1];
char line_op[NBL+1];
char line_d2[NBL+1];
char line_dpos2[NBL+1];



void flush()
{
  int i;
  if (line_pos > 0){ 
    /* close strings */
    i = line_pos;
    if (line_dpos1[i]!='\0' ){
      while (line_dpos1[i]!=' ' && i>0)
	i--;
      line_dpos1[i]='\0';
    }
    
    i = line_pos;
    if (line_dpos2[i]!='\0'){
      while (line_dpos2[i]!=' ' && i>0)
	i--;
      line_dpos2[i]='\0';
    }
			       
    line_dpos1[line_pos] = '\0';
    line_dpos2[line_pos] = '\0';
    line_d1[line_pos]    = '\0';
    line_d2[line_pos]    = '\0';
    line_op[line_pos]    = '\0';
    /* print the 5 lines */
    fprintf(OUTSTREAM,line_dpos1);fprintf(OUTSTREAM,"\n");
    fprintf(OUTSTREAM,line_d1   );fprintf(OUTSTREAM,"\n");
    fprintf(OUTSTREAM,line_op   );fprintf(OUTSTREAM,"\n");
    fprintf(OUTSTREAM,line_d2   );fprintf(OUTSTREAM,"\n");
    fprintf(OUTSTREAM,line_dpos2);fprintf(OUTSTREAM,"\n");
    fprintf(OUTSTREAM,"\n");
    /* reset buffers */
    line_pos = 0;
    memset(line_dpos1,0,NBL+1);
    memset(line_dpos2,0,NBL+1);
    memset(line_d1   ,0,NBL+1);
    memset(line_d2   ,0,NBL+1);
    memset(line_op   ,0,NBL+1);
  }
}

void push(char d1,int dpos1,char op,char d2, int dpos2)
{
  /* put positions in the buffer */
  if ( !(dpos1%10)  && size_int(dpos1) + line_pos < NBL){
    if (line_pos>0)
      if (line_dpos1[line_pos-1] == '|')
	line_dpos1[line_pos-1] = ' ';
      sprintf(line_dpos1+line_pos,  "|%i",dpos1);
  }
  if (line_dpos1[line_pos]=='\0')
    line_dpos1[line_pos]=' ';
  
  
  if (!(dpos2%10)   && size_int(dpos2) + line_pos < NBL){
    if (line_pos>0)
      if (line_dpos2[line_pos-1] == '|')
	line_dpos2[line_pos-1] = ' ';
    sprintf(line_dpos2+line_pos,"|%i",dpos2);
  }
  if (line_dpos2[line_pos]=='\0')
    line_dpos2[line_pos]=' ';
 
  
  /* memorize alignment */
  
  line_d1[line_pos]=d1;
  line_d2[line_pos]=d2;  
  line_op[line_pos]=op;
  
  line_pos++;

  /* print if end of line */
  if (line_pos == NBL){
    flush();
  } 
}


/*************************************************************/





/* SIMPLE ALIGNEMENT 
 * $ GAP1
 * $ SUB1
 */


#define M(i,j) (mt[(i)+ (j)*(len1+1)])
#define D(k)   (bk[k])

int display_alignment_noflush(int reverse,
			      char * data1, int pos1, int len1,
			      char * data2, int pos2, int len2)
{
  
  int i,j,diag;
  int i1,i2;
  int k;
  int * mt;
  backtrack * bk;

  mt = (int*) MALLOC( (unsigned) (len1+1) * (len2+1) * sizeof(int));
  ASSERT(mt,display_alignment_noflush);
  bk = (backtrack*) MALLOC( (unsigned) (len1+len2+1) * sizeof(backtrack));
  ASSERT(bk,display_alignment_noflush);

  /* [1] compute */
  M(0,0) = 0;
  
  for(j=0; j < len2; j++)
    M(0,j+1) = M(0,j) + GAP1; 
  
  for(i=0; i < len1; i++)                           
    { 
      M(i+1,0) = M(i,0)  + GAP1; 
      for(j=0; j < len2 ; j++)                        
	{ 
	  diag = M(i,j);
	  if( TNORM(data1[pos1+i]) != TNORM(data2[pos2+j])) 
	    diag+=SUB1;
	  M(i+1,j+1) = MAX3 ( diag, M(i,j+1) + GAP1 , M(i+1,j) + GAP1);
	  
	}		
    }	
  
  
  /* [2] backtrace */ 
  i = len1;
  j = len2;
  k = 0;
  
  while (i > 0 && j > 0){
      if ( M(i,j) ==  M(i-1,j)  + GAP1 ){
	i--;
	D(k++) = STEP_LEFT;
      }else{ 
	if ( M(i,j) ==  M(i,j-1) + GAP1){
	  j--;
	  D(k++) = STEP_RIGHT;
	}else{
	  if ( M(i,j) == M(i-1,j-1) && (TNORM(data1[pos1+i-1]) == TNORM(data2[pos2+j-1]))){ 
	    i--;
	    j--;
	    D(k++) = DIAG_EGAL;
	  }else{ 
	    if ( M(i,j) ==  M(i-1,j-1) + SUB1 && (TNORM(data1[pos1+i-1]) != TNORM(data2[pos2+j-1]))){
	      j--;i--;
	      D(k++) = DIAG_DIFF;
	    }else{
	      _WARNING("display_alignment_noflush() 1");
	    }
	  }
	}
      }
  }
  
  while(j > 0){
    j--;
    D(k++) = STEP_RIGHT;
  }
  
  while(i > 0){
    i--;
    D(k++) = STEP_LEFT;
  }
  
  FREE(mt, (unsigned) (len1+len2+1) * sizeof(int));
  
  /* [3] display alignment */

  i1 = 0; i2 = 0; k--;
  
  while (i1 < len1 || i2 < len2 ) {
    
    switch(D(k--)){
    case DIAG_EGAL:
      push(LOOKUP(data1[pos1+i1]),
	   BCOUNT + ((reverse) ? (gp_querysize_chunk - 1) - (pos1 + i1) : pos1+i1),
	   CHAR_EQU,
	   LOOKUP(data2[pos2+i2]),
	   BCOUNT + pos2 + i2
	   );
      i1++;i2++;
      break;
	
    case DIAG_DIFF:
      push(LOOKUP(data1[pos1+i1]),
	   BCOUNT + ((reverse) ? (gp_querysize_chunk - 1)- (pos1 - i1) : pos1+i1),
	   (((TNORM(data1[pos1+i1]))^(TNORM(data2[pos2+i2])))==2)?CHAR_SS:CHAR_SV,
	   LOOKUP(data2[pos2+i2]),
	   BCOUNT + pos2 + i2
	   );
      
      i1++;i2++;
      break;
      
    case STEP_LEFT:
      push(LOOKUP(data1[pos1+i1]),
	   BCOUNT + ((reverse) ? (gp_querysize_chunk - 1) - (pos1 + i1) : pos1+i1),
	   CHAR_DEL,
	   CHAR_DUM,
	   BCOUNT + pos2 + i2
	   );
      i1++;
      break;
      
    case STEP_RIGHT:
      push(CHAR_DUM,
	   BCOUNT + ((reverse) ? (gp_querysize_chunk - 1) - (pos1 + i1) : pos1+i1),
	   CHAR_INS,
	   LOOKUP(data2[pos2+i2]),
	   BCOUNT + pos2+i2
	   );
      i2++;
      break;
      
    default:
      _WARNING("display_alignment_noflush() 2");
    }
    
  }/* while */

  FREE(bk,(unsigned) (len1+len2+1) * sizeof(backtrack));
  return 0;
}



 int display_alignment(int reverse,
			     char * data1, int pos1, int len1,
			     char * data2, int pos2, int len2)
{
  display_alignment_noflush(reverse,
			    data1,pos1,len1,
			    data2,pos2,len2);
  flush();
  return 0;  
}


 int display_alignment_on_MA(char * data1, char * data2,
				   MA * ma)
     
{
  int i;
  int left_size,right_size;
  int left_correction=0;
  tuple * t = ma->first_tuple, * t_prev = ma->first_tuple;

  /* [1] Display alignment before first tuple */
  left_size  = TBL_POS(t) - ma->left_pos_begin  ;
  right_size = TBR_POS(t) - ma->right_pos_begin ;
  
  if ( left_size > 0 && right_size > 0)
    display_alignment_noflush(ma->reverse,
			      data1, ma->left_pos_begin , left_size ,
			      data2, ma->right_pos_begin, right_size
			      );

  /* [2] Display alignment for several tuples */
  while(t != NULL){

    /* (1) display alignment of tuple "t" */
    for(i = 0 ;i < TSIZE(t) ; i++){
      push(LOOKUP(data1[i+TBL_POS(t)]),
	   BCOUNT + ((ma->reverse) ? (gp_querysize_chunk - 1) - (i + TBL_POS(t)) : i+TBL_POS(t)),
	   (TNORM(data1[i+TBL_POS(t)]) == TNORM(data2[i+TBR_POS(t)]))?CHAR_EQU:((TNORM(data1[i+TBL_POS(t)])^TNORM(data2[i+TBR_POS(t)]))==2)?CHAR_SS:CHAR_SV,
	   LOOKUP(data2[i+TBR_POS(t)]),
	   BCOUNT + i+TBR_POS(t)
	   );      
    }

    
    if (t->next){
      /* (2) display alignment between (t) and (t->next) */
      display_alignment_noflush(ma->reverse,
				data1,TEL_POS(t),TGAP_L(t,t->next),
				data2,TER_POS(t),TGAP_R(t,t->next));
    }
    t = t->next;
    t_prev = t;
  }

  /* [3] Display alignment after first tuple */
  
  left_size  =  ma->left_pos_end  - TEL_POS(t_prev) ;
  right_size =  ma->right_pos_end - TER_POS(t_prev) ;

  if ( left_size > 0 && right_size> 0)
    display_alignment_noflush(ma->reverse,
			      data1, TEL_POS(t_prev), left_size ,
			      data2, TER_POS(t_prev), right_size
			      );
  flush();
 
  return 0;
}


/* ALIGNEMENT WITh SUBS MATRIX + GAP OPEN CONTINUE
 * gp_cost_gap_opened
 * gp_cost_gap_continued
 * substitution_matrix
 */



 int display_alignment_SG_noflush(int reverse,
					char * data1, int pos1, int len1,
					char * data2, int pos2, int len2)
{
  
  int i,j;
  int i1,i2;
  int k;
  int * mt, * in; 
  backtrack_SG * bk;
  
  mt = (int*) MALLOC( (unsigned) (len1+1) * (len2+1) * sizeof(int));
  ASSERT(mt,display_alignment_SG_noflush);
  in = (int*) MALLOC( (unsigned) (len1+1) * (len2+1) * sizeof(int));  
  ASSERT(in,display_alignment_SG_noflush);
  bk = (backtrack_SG*) MALLOC( (unsigned) (len1+len2+1) * sizeof(backtrack_SG));
  ASSERT(bk,display_alignment_SG_noflush);
  

  /* [1] compute */
  
  M3(0,0) = 0; 
  I3(0,0) = gp_cost_gap_opened - gp_cost_gap_continued;
  
  for(i=1 ; i<len1+1 ; i++)
    M3(i,0) = -INFINITY_INT;
 
  for(i=1 ; i<len1+1 ; i++)
    I3(i,0) = I3(i-1,0) +  gp_cost_gap_continued;
  
  for(j=1 ; j<len2+1 ; j++){
    
    M3(0,j) =  -INFINITY_INT;
    I3(0,j) =  I3(0,j-1)  + gp_cost_gap_continued;
    
    for(i=1;i<len1+1;i++){
      I3(i,j) = MAX (
		     MAX( M3(i,j-1) + gp_cost_gap_opened    , M3(i-1,j) + gp_cost_gap_opened )
		     ,
		     MAX( I3(i,j-1) + gp_cost_gap_continued , I3(i-1,j) + gp_cost_gap_continued )
		     );
      if  (data1 + pos1 + i !=  data2 + pos2 + j)
	M3(i,j) = MAX ( M3(i-1,j-1) + S3(i,j), I3(i-1,j-1) + S3(i,j) ); 
      else
	M3(i,j) = -INFINITY_INT;
    }
  }
 

#ifdef DEBUGDISPLAYALIGN
  {  
    int i,j;
    fprintf(stdout,"   ");
    for (i=0; i<=len1; i++)
      fprintf(stdout,"|  %c  ",lookup[(int)data1[pos1+i]]);
    fprintf(stdout,"|\n");
    for (j=0; j<=len2; j++) {
      fprintf(stdout," %c |",lookup[(int)data2[pos2+j]]);
      for (i=0; i<=len1; i++) {
	fprintf(stdout,"%2.1g|",(float)(M3(i,j)));    
      }
      fprintf(stdout,"\n");
    }
  }
#endif

  /* [2] backtrace */ 
  i = len1;
  j = len2;
  k = 0;
  
  if (M3(len1,len2) < I3(len1,len2))
    D3(k++) = I_TO_M_SUBS;
  else
    D3(k++) = M_TO_M_SUBS;
  
  
  while (i > 0 && j > 0){
        
    if ( D3(k-1) == M_TO_M_SUBS || D3(k-1) == FIRST_DEL_LEFT || D3(k-1) == FIRST_DEL_RIGHT ){

      /* no gap to open */

      if ( M3(i,j) == M3(i-1,j-1) + S3(i,j) ){ 
	i--;
	j--;
	D3(k++) = M_TO_M_SUBS;
	}else{	
	  if ( M3(i,j) == I3(i-1,j-1) + S3(i,j) ){ 
	    i--;
	    j--;
	    D3(k++) = I_TO_M_SUBS;
	  }else{
	    _WARNING("display_alignment_SG_noflush() 1");
	  }
	}
    }else{
      
      /* need to open a gap */
      if ( I3(i,j) == M3(i,j-1) + gp_cost_gap_opened ){
	j--;
	D3(k++) =  FIRST_DEL_RIGHT ;
      }else{
	if ( I3(i,j) == M3(i-1,j) + gp_cost_gap_opened ){
	  i--;
	  D3(k++) =  FIRST_DEL_LEFT ;
	}else{
	  if ( I3(i,j) == I3(i,j-1) + gp_cost_gap_continued ){
	    j--;
	    D3(k++) =  CONT_DEL_RIGHT ;
	  }else{
	    if ( I3(i,j) == I3(i-1,j) + gp_cost_gap_continued ){
	      i--;
	      D3(k++) =  CONT_DEL_LEFT ;
	    }else{
	      _WARNING("display_alignment_SG_noflush() 2");
	    }
	  } 
	}
      }
    }
  }/* while */
  
  if (j > 0) {
    while(j > 1) {
      j--;
      D3(k++) = CONT_DEL_RIGHT;
    }  
    D3(k++) = FIRST_DEL_RIGHT;
  }
  
  if (i > 0) {
    while(i > 1) {
      i--;
      D3(k++) = CONT_DEL_LEFT;
    }
    D3(k++) = FIRST_DEL_LEFT;
  }
  
  FREE(mt,(unsigned) (len1+1) * (len2+1) * sizeof(int));
  FREE(in,(unsigned) (len1+1) * (len2+1) * sizeof(int));


#ifdef DEBUGDISPLAYALIGN
  fprintf(stdout,"backtrack : ");
  for( i = k ; i >= 0 ; i--){
    fprintf(stdout,"[%d]",D3(k));
  }
  fprintf(stdout,"\n");
#endif


  /* [3] display alignment */
 
  i1 = 0; i2 = 0; k-- ;
 
  while (i1 < len1 || i2 < len2 ){
    
    switch(D3(k--)){
    case M_TO_M_SUBS:
    case I_TO_M_SUBS:
      push(LOOKUP(data1[pos1+i1]),
	   BCOUNT + ((reverse) ? (gp_querysize_chunk - 1) - (pos1 + i1) : pos1+i1),
	   (TNORM(data1[pos1+i1])==TNORM(data2[pos2+i2]))?CHAR_EQU:(((TNORM(data1[pos1+i1])^TNORM(data2[pos2+i2]))==2)?CHAR_SS:CHAR_SV),
	   LOOKUP(data2[pos2+i2]),
	   BCOUNT + pos2+i2
	   );
      i1++;
      i2++;
    break;
	
    case FIRST_DEL_LEFT:
    case CONT_DEL_LEFT:
      push(LOOKUP(data1[pos1+i1]),
	   BCOUNT + ((reverse) ? (gp_querysize_chunk - 1) - (pos1 + i1) : pos1+i1),
	   CHAR_DEL,
	   CHAR_DUM,
	   BCOUNT + pos2+i2
	   );
      i1++;
      break;
      
    case FIRST_DEL_RIGHT:
    case CONT_DEL_RIGHT:
      push(CHAR_DUM,
	   BCOUNT + ((reverse) ? (gp_querysize_chunk - 1) - (pos1 + i1) : pos1+i1),
	   CHAR_INS,
	   LOOKUP(data2[pos2+i2]),
	   BCOUNT + pos2+i2
	   );
      i2++;
      break;
      
    default:
      _WARNING("display_alignment_SG_noflush() 3");
    } 
    
  }/* while */
  FREE(bk,(unsigned)(len1+len2+1) * sizeof(backtrack_SG));

  return 0;  
}





 int display_alignment_SG(int reverse,
			  char * data1, int pos1, int len1,
			  char * data2, int pos2, int len2)
{
  display_alignment_SG_noflush(reverse,
			       data1,pos1,len1,
			       data2,pos2,len2);
  flush();
  return 0;  
}


int display_alignment_SG_on_MA(char * data1,char * data2,
				MA * ma)
			    
{
  int i;
  int left_size,right_size;
  int left_correction=0;
  tuple * t = ma->first_tuple, * t_prev = ma->first_tuple;
  
  /* [1] Display alignment before first tuple */
  
  left_size  = TBL_POS(t) - ma->left_pos_begin  ;
  right_size = TBR_POS(t) - ma->right_pos_begin ;

#ifdef DEBUG_DISPLAY
#define DEBUG_DISPLAY_PADDING 50
  display_alignment_SG_noflush(ma->reverse,
			       data1, MAX(0, ma->left_pos_begin  - DEBUG_DISPLAY_PADDING) , MIN(DEBUG_DISPLAY_PADDING,ma->left_pos_begin),
			       data2, MAX(0, ma->right_pos_begin - DEBUG_DISPLAY_PADDING) , MIN(DEBUG_DISPLAY_PADDING,ma->right_pos_begin));  
  push('{',1,'{','{',1); 
#endif  
  
  if ( left_size > 0 && right_size > 0)
    display_alignment_SG_noflush(ma->reverse,
			      data1, ma->left_pos_begin , left_size ,
				 data2, ma->right_pos_begin, right_size
			      );
 
#ifdef DEBUG_DISPLAY
  push('<',1,'<','<',1); 
#endif 

  /* [2] Display alignment for several tuples */
  
  while(t != NULL){
    
    /* (1) display alignment of tuple "t" */     
    for(i = 0 ;i < TSIZE(t) ; i++){
      push(LOOKUP(data1[i+TBL_POS(t)]),
	   BCOUNT + ((ma->reverse) ? (gp_querysize_chunk - 1) - (i + TBL_POS(t)) : i+TBL_POS(t)),
	   (TNORM(data1[i+TBL_POS(t)]) == TNORM(data2[i+TBR_POS(t)]))?CHAR_EQU:((TNORM(data1[i+TBL_POS(t)])^TNORM(data2[i+TBR_POS(t)]))==2)?CHAR_SS:CHAR_SV,
	   LOOKUP(data2[i+TBR_POS(t)]),
	   BCOUNT + i + TBR_POS(t)
	   );      
    }

    
    if (t->next){
      /* (2) display alignment between (t) and (t->next) */
      display_alignment_SG_noflush(ma->reverse,
				   data1,TEL_POS(t),TGAP_L(t,t->next),
				   data2,TER_POS(t),TGAP_R(t,t->next));
    }
    t_prev = t;
    t = t->next;
  }
  

#ifdef DEBUG_DISPLAY
  push('>',1,'>','>',1); 
#endif

  /* [3] Display alignment after first tuple */
 
  left_size  =  ma->left_pos_end  - TEL_POS(t_prev) ;
  right_size =  ma->right_pos_end - TER_POS(t_prev) ;
  
  if ( left_size > 0 && right_size> 0)
    display_alignment_SG_noflush(ma->reverse,
				 data1, TEL_POS(t_prev), left_size ,
				 data2, TER_POS(t_prev), right_size);
 
#ifdef DEBUG_DISPLAY    
  push('}',1,'}','}',1);  
  display_alignment_SG_noflush(ma->reverse,
			       data1, TEL_POS(t_prev)+left_size,  MIN(DEBUG_DISPLAY_PADDING, gp_querysize_chunk             - TEL_POS(t_prev) - left_size),
			       data2, TER_POS(t_prev)+right_size, MIN(DEBUG_DISPLAY_PADDING, gp_chunksize_text[ma->i_chunk] - TER_POS(t_prev) - right_size));
#endif
  flush();
  return 0;
}




/*******************************
 *
 *  Compute the real score
 *
 ********************************/



int alignment_score_on_MAs(char * data1, char * data2,MA * firstMa,int left_correction){
  MA * ma=firstMa;
  
  while(ma!=NULL) {
    alignment_SG_score_on_MA(data1,data2,ma,left_correction);
    /* fprintf(stderr,"(%d,%d)(%d,%d)  score=%d\n",ma->left_pos_begin,ma->left_pos_end,ma->right_pos_begin,ma->right_pos_end,ma->blastscore); */
    ma=ma->next;
  }
  return 1;
}




int alignment_SG_score_on_MA(char * data1, char * data2,
			     MA * ma,int left_correction){
  int left_size,right_size;
  tuple * t = ma->first_tuple, * t_prev = ma->first_tuple ;
  ma->blastscore=0;
  
  /* [1] Compute alignment before first tuple */  
  left_size  = TBL_POS(t) - ma->left_pos_begin  ;
  right_size = TBR_POS(t) - ma->right_pos_begin ;
  
  if ( left_size > 0 && right_size > 0)
    alignment_SG_score(data1, ma->left_pos_begin , left_size ,
		       data2, ma->right_pos_begin, right_size,
		       &(ma->blastscore));
  /* fprintf(stderr,"score a gauche %d\n",ma->blastscore); */
  /* [2] Compute alignment for several tuples */
 
  while(t != NULL){
    /* (1) compute inside a tuple */
    alignment_SG_score_Strait(data1,TBL_POS(t),
			      data2,TBR_POS(t),TSIZE(t),
			      &(ma->blastscore));
    /*fprintf(stderr,"score apres strait: %d length=%d (%d-%d)(%d-%d)\n",ma->blastscore,TSIZE(t),
	    TBL_POS(t),TEL_POS(t),
	    TBR_POS(t),TER_POS(t));*/
    if (t->next){
      /* (2) alignment between (t) et (t->next) */
      alignment_SG_score(data1,TEL_POS(t),TGAP_L(t,t->next),
			 data2,TER_POS(t),TGAP_R(t,t->next),
			 &(ma->blastscore));
      /*fprintf(stderr,"score entre tuple: %d\n",ma->blastscore);*/
    }/* if */
    t_prev = t;
    t = t->next;
  }/* while */

  /* [3] Compute alignement after last tuple */
  
  left_size  =  ma->left_pos_end  - TEL_POS(t_prev) ;
  right_size =  ma->right_pos_end - TER_POS(t_prev) ;
  /*fprintf(stderr,"score a avant  droite:%d len1=%d len2=%d\n",ma->blastscore,left_size,right_size);*/
  
  
  if ( left_size > 0 && right_size> 0)
    alignment_SG_score(data1, TEL_POS(t_prev), left_size ,
		       data2, TER_POS(t_prev), right_size,
		       &(ma->blastscore));
  /*printf(stderr,"score apres  droite:%d\n",ma->blastscore);*/
  return 0;
}


int alignment_SG_score_Strait(char * data1,  int pos1, char * data2,  int pos2, int len,
			      /*out*/int    *p_score){
  int i;

  for(i=0;i<len;i++){ 
     *p_score=*p_score-gp_substitution_matrix[((int)(data1 + pos1)[i])][(int)(data2 + pos2)[i]];
     /*fprintf(stderr,"STRAIT: %c<->%c score=%d\n",lookup[data1[pos1+i]],lookup[data2[pos2+i]],
       gp_substitution_matrix[((int)(data1 + pos1)[i])][(int)(data2 + pos2)[i]]); */

  }
  return 0;
}
 



int alignment_SG_score(char * data1, int pos1, int len1,
		       char * data2, int pos2, int len2,
		       /*out*/int    *p_score
		       ){
  int i,j;
  int gain_score;
  int * mt, * in;
  mt = (int*) MALLOC( (unsigned) (len1+1) * (len2+1) * sizeof(int));
  ASSERT(mt,alignment_SG_score);
  in = (int*) MALLOC( (unsigned) (len1+1) * (len2+1) * sizeof(int));  
  ASSERT(in,alignment_SG_score);
  

  M3(0,0) = 0;
  
  I3(0,0) = gp_cost_gap_opened - gp_cost_gap_continued;
  
  for(i=1;i<len1+1; i++)
    M3(i,0) = -INFINITY_INT;
  
  for(i=1 ; i<len1+1 ; i++)
    I3(i,0) = I3(i-1,0) + gp_cost_gap_continued;
  

  for(j=1 ; j<len2+1 ; j++){
  
    M3(0,j) =  -INFINITY_INT;
    I3(0,j) =  I3(0,j-1)  + gp_cost_gap_continued;
    
    for(i=1;i<len1+1;i++){
      I3(i,j) = MAX (
		     MAX( M3(i,j-1) + gp_cost_gap_opened    , M3(i-1,j) + gp_cost_gap_opened )
		     ,
		     MAX( I3(i,j-1) + gp_cost_gap_continued , I3(i-1,j) + gp_cost_gap_continued )
		     );
      M3(i,j) = MAX ( M3(i-1,j-1) + S3(i,j), I3(i-1,j-1) + S3(i,j) ); 
    }
  }
  
  gain_score = MAX( M3(len1,len2) , I3(len1,len2) );
  *p_score += gain_score;
  return gain_score; 
}
