/*
 *  YASS 1.11
 *  Copyright (C) 2004-2007
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "global_var.h"
#include "tuple.h"
#include "prdyn.h"
#include "proba.h"
#include "kword.h"


#ifdef INLINE
inline
#endif
tuple * CreateTuple(int occurrence, int diagonal, int seed)
{

    tuple *t = (tuple *) MALLOC((unsigned) sizeof(tuple));
    ASSERT(t, CreateTuple);
    t->occurrence = occurrence;
    t->diagonal = diagonal;
    t->leftsize = 0;
    t->bitweight = gp_seeds_bitweight[seed];
    t->next = NULL;
    t->lastSeed = seed;
    t->overlappedhits = gp_seeds_bitmask[seed];
    return t;
}


#ifdef INLINE
inline
#endif
tuplelist * CreateTupleList()
{
  tuplelist *tl = (tuplelist *) MALLOC((unsigned) sizeof(tuplelist));
  ASSERT(tl, CreateTupleList);
  tl->first_tuple = NULL;
  tl->next = NULL;
  return tl;
}



#ifdef INLINE
inline
#endif
void KeepTupleList(tuplelist * tl, tuple * tfirst, tuple * tlast, MA * ma)
{
    tuple *t = tl->first_tuple->next;
    tuple *t_prev = tl->first_tuple;
    if (tfirst == tl->first_tuple) {
	tl->first_tuple = tlast->next;
    } else {
	while (t != tfirst) {
	    t_prev = t;
	    t = t->next;
	}
	t_prev->next = tlast->next;
    }
    ma->first_tuple = tfirst;
    tlast->next = NULL;
}



#ifdef INLINE
inline
#endif
void FreeTupleList(tuplelist ** p_tl, tuplelist ** p_tl_prev, tuplelist **last_tl)
{
    tuple *t, *t_prev;

    if (*last_tl == *(p_tl))
	*last_tl = *(p_tl_prev);

    t = (*(p_tl))->first_tuple;
    while (t != NULL) {
	t_prev = t;
	t = t->next;
	FREE(t_prev,sizeof(tuple));
    }

    (*(p_tl_prev))->next = (*(p_tl))->next;
    FREE(*(p_tl),sizeof(tuplelist));

    *(p_tl) = (*(p_tl_prev))->next;
}



#ifdef INLINE
inline
#endif
void FreeTuples(tuple * t)
{
  tuple * t_ =  t, * t_prev_;
  while (t_ != NULL) {
    t_prev_ = t_;
    t_ = t_->next;
    FREE(t_prev_,sizeof(tuple));
  }
}



#ifdef INLINE
inline
#endif
MA * CreateMA(int left_pos_begin, int right_pos_begin,
		  int left_pos_end, int right_pos_end,
		  tuplelist * tuple_list,
		  tuple * first_tuple_aligned,
		  tuple * last_tuple_aligned, 
	          int blastscore, double entropy,
		  Feature * feature
		  )
{

    MA * ma = NULL;
    
    ma = MALLOC((unsigned) sizeof(MA));
    ASSERT(ma, memorize_alignment);
    
    /* 1) initialise ma */
    ma->next = NULL;
    ma->left_pos_begin = left_pos_begin;
    ma->right_pos_begin = right_pos_begin;
    ma->left_pos_end = left_pos_end;
    ma->right_pos_end = right_pos_end;
    ma->blastscore = blastscore;


    
    
    /* 2) keep the seeds chain inside MA */
    KeepTupleList(tuple_list, first_tuple_aligned, last_tuple_aligned, ma);

    ma->reverse = (char) feature->reverse;
    ma->i_chunk = feature->i_chunk;
    
    /* 3) add  "ma" in the MA_linked_list */
    if (feature->first_MA == NULL) {
	feature->first_MA = ma;
    } else {
	feature->last_MA->next = ma;
    } 
    feature->last_MA = ma;
   
#ifdef STATS	
    feature->nb_ma++;
#endif
    

    /* left_correction take effect on TL_POS(t): we can do alignments before calling the "left_correction"
     * function on each MA, while left_correction global variable is != 0, or either after the "left_correction()"
     * call provided that left_correction global variable IS = 0.
     */

    /* correction due to the algorithm "Assemble_Double" over "Assemble_Single" */
    
    LeftCorrection_MA(ma,feature->left_correction);

    return ma;
}

#ifdef INLINE
inline
#endif
void FreeMA(MA * ma)
{
    tuple *t_, *t_prev_;
    t_ = ma->first_tuple;
    while (t_) {
	t_prev_ = t_;
	t_ = t_->next;
	FREE(t_prev_,sizeof(tuple));
    }
    FREE(ma,sizeof(MA));
}

#ifdef INLINE
inline
#endif
void LeftCorrection_MA(MA * ma,int left_correction)
{
    tuple *t = ma->first_tuple;
    while (t) {
	t->diagonal -= left_correction;
	t = t->next;
    }
}

/*

int QuickSort_MAList(MA ** p_firstMA, MA ** p_lastMA, SortFunct * f )
{


  MA * _ma_top    = *(p_firstMA);
  MA * _ma_1      = NULL;
  MA * _ma_1_top  = NULL;
  MA * _ma_0      = NULL;
  MA * _ma_0_top  = NULL;
  MA * _ma        = NULL;
  
  if (_ma_top) {
    
    _ma = _ma_top->next;
  
    while (_ma) {
      if (f(_ma,_ma_top)) {
	if (!_ma_1)
	  _ma_1 = _ma;
	else
	  _ma_1_top->next = _ma;
	_ma_1_top = _ma;
      } else {
	if (!_ma_0)
	  _ma_0 = _ma;
	else
	  _ma_0_top->next = _ma;
	_ma_0_top = _ma;
      }
      _ma = _ma->next;
    }
  
  
    if (_ma_0_top) {
      _ma_0_top->next = NULL;
      if (_ma_0 != _ma_0_top)
	QuickSort_MAList(&_ma_0,&_ma_0_top,f);
    }

    if (_ma_1_top) {
      _ma_1_top->next = NULL;
      if (_ma_1 != _ma_1_top)
	QuickSort_MAList(&_ma_1,&_ma_1_top,f);
    }



    
    _ma_top->next =_ma_0;
    if (_ma_1_top) {
      _ma_1_top->next = _ma_top;
      *(p_firstMA) = _ma_1;
    } else {
      *(p_firstMA) = _ma_top;
    }
    
    if (p_lastMA) {
      if (_ma_0_top)
	*(p_lastMA) = _ma_0_top;
      else
	*(p_lastMA) = _ma_top;
    }

    if (*(p_lastMA))
      (*(p_lastMA))->next = NULL;
  }  
  return 0;
}

*/


/* ============================================================================ */

int SortCriterionScore(MA * ma)            { return (ma->blastscore);  }
int SortCriterionEntropy(MA * ma)          { return (int)((ma->entropy)*1000);}
int SortCriterionMutual(MA * ma)           { return (int)((ma->mutual)*1000);}
int SortCriterionScoreWithEntropy(MA * ma) { return (int)((ma->entropy) * (ma->blastscore));}
int SortCriterionQueryBegin(MA * ma)       { return (int)((ma->reverse) ? 
					                0x7fffffff-(ma->left_pos_begin):
							0x7fffffff-(gp_querysize_chunk -( ma->left_pos_end)));
 
					       }



/*
 * ListSort
 */

void ListSort_MAList(MA ** p_firstMA, MA ** p_lastMA, int mask /* 0x40000000 */ )
{

    MA *ma_0 = NULL, *ma_1 = NULL;
    MA *ma_0_top = NULL, *ma_1_top = NULL;

    MA *_ma = *(p_firstMA);

    while (_ma) {
	if ((gp_sortcriterion_func(_ma)) & mask) {
	    if (!ma_1)
		ma_1 = _ma;
	    else
		ma_1_top->next = _ma;
	    ma_1_top = _ma;
	} else {
	    if (!ma_0)
		ma_0 = _ma;
	    else
		ma_0_top->next = _ma;
	    ma_0_top = _ma;
	}
	_ma = _ma->next;
    }

    mask >>= 1;

    if (ma_0_top) {
	ma_0_top->next = NULL;
	if (mask > 0)
	    ListSort_MAList(&ma_0, &ma_0_top, mask);
    }
    if (ma_1_top) {
	ma_1_top->next = NULL;
	if (mask > 0)
	    ListSort_MAList(&ma_1, &ma_1_top, mask);
    }


    /* chain the two sublists into one */
    if (ma_1_top) {
	*(p_firstMA) = ma_1;
	if (ma_0_top) {
	    ma_1_top->next = ma_0;
	    *(p_lastMA) = ma_0_top;
	} else {
	    *(p_lastMA) = ma_1_top;
	}
    } else {
	*(p_firstMA) = ma_0;
	*(p_lastMA)  = ma_0_top;

    }
    /* lastMA->next is set to NULL */
    if (*(p_lastMA))
      (*(p_lastMA))->next = NULL;
}





/*
 * main function used to sort MA 
 */

void Sort_MAList(Feature * f) {
  
  if (f->first_MA) {
    
    if (gp_sortcriterion >= NBSORTCRITERIA) {
      
      /* [A] Sort with chunk selection */
      MA * _ma_after_end    = f->first_MA; 
      MA * _ma_before_start = NULL;
      
      while(_ma_after_end) {
	MA *  _ma_end   = _ma_after_end;
	MA *  _ma_start = _ma_after_end;
	int   i_chunk   = _ma_after_end->i_chunk; 
	
	/* search sublist 's' with same chunk */
	while(_ma_after_end && (_ma_after_end->i_chunk ==  i_chunk)) {
	  _ma_end        = _ma_after_end;
	  _ma_after_end  = _ma_after_end->next;
	}
	
	/* sort and insert this sublist 's'*/
	if (_ma_start != _ma_end){
	  /* sort */
	  _ma_end->next = NULL;
	  ListSort_MAList(&_ma_start,&_ma_end,0x40000000);
	  _ma_end->next    = _ma_after_end;
	  /* insert */
	  if (!_ma_before_start)
	    f->first_MA            = _ma_start;
	  else
	    _ma_before_start->next = _ma_start;
	  _ma_before_start = _ma_end;
	}
	
	f->last_MA = _ma_end;
      }
      
    } else {
      /* [B] Sort without chunk selection */
      ListSort_MAList(&(f->first_MA),&(f->last_MA),0x40000000);
    }
  }
}


/*
 * main function used to sort MA 
 */

MA *  MergeSort_MALists(Feature * f1, Feature * f2) {
  MA * first_MA = NULL;
  MA * last_MA  = NULL;

  if (f1->first_MA && f2->first_MA) {
    
    if (gp_sortcriterion >= NBSORTCRITERIA) {

      /* [A] MergeSort with chunk selection */
       
      MA * _ma_1 = f1->first_MA;
      MA * _ma_2 = f2->first_MA;
      
      if ((_ma_1 > _ma_2) || ((_ma_1 == _ma_2) && (gp_sortcriterion_func(_ma_1) > gp_sortcriterion_func(_ma_2)))) {
	first_MA = last_MA = _ma_1;
	_ma_1    = _ma_1->next;
      } else  {
	first_MA = last_MA = _ma_2;
	_ma_2    = _ma_2->next;
      }
      
      while(_ma_1 && _ma_2) {
	if ((_ma_1 > _ma_2) || ((_ma_1 == _ma_2) && (gp_sortcriterion_func(_ma_1) > gp_sortcriterion_func(_ma_2)))) {
	  last_MA->next = _ma_1;
	  last_MA       = _ma_1;
	  _ma_1    = _ma_1->next;
	} else  {
	  last_MA->next = _ma_2;
	  last_MA       = _ma_2;
	  _ma_2    = _ma_2->next;
	}
      }
	
      if (!_ma_1){
	last_MA->next = _ma_2;
      }else{ 
	if (!_ma_2){
	  last_MA->next = _ma_1;
	} else {
	  last_MA->next = NULL;
	}
      }
      
     
    } else {

      /* [B] MergeSort without chunk selection */
       
      MA * _ma_1 = f1->first_MA;
      MA * _ma_2 = f2->first_MA;
      
      if (gp_sortcriterion_func(_ma_1) > gp_sortcriterion_func(_ma_2)) {
	first_MA = last_MA = _ma_1;
	_ma_1    = _ma_1->next;
      } else  {
	first_MA = last_MA = _ma_2;
	_ma_2    = _ma_2->next;
      }
      
      while(_ma_1 && _ma_2) {
	if (gp_sortcriterion_func(_ma_1) > gp_sortcriterion_func(_ma_2)) {
	  last_MA->next = _ma_1;
	  last_MA       = _ma_1;
	  _ma_1    = _ma_1->next;
	} else  {
	  last_MA->next = _ma_2;
	  last_MA       = _ma_2;
	  _ma_2    = _ma_2->next;
	}
      }
	
      if (!_ma_1){
	last_MA->next = _ma_2;
      }else{ 
	if (!_ma_2){
	  last_MA->next = _ma_1;
	} else {
	  last_MA->next = NULL;
	}
      }

    }
  } else {
     if (f1->first_MA)
       first_MA = f1->first_MA;
     else
       first_MA = f2->first_MA;
  }
  
  return first_MA;
}



/*
 * return true if the filter does apply on the current MA
 * and thus the MA has to be removed
 */

int EntropyAndScoreFilter_MA(Feature * f, MA * ma, char *dataquery,  char *datachunktext)
{
  /* Score filter */
  if (ma->blastscore < f->MAminscore)
    return 1;
  
  /* Entropy filter */
  memset(gv_nb_mutations_per_codon1, '\0', 3 * sizeof(int));
  memset(gv_nb_mutations_per_codon2, '\0', 3 * sizeof(int));
  memset(gv_nb_transitions_transversions_indels, '\0', 3 * sizeof(int));
  memset(gv_nb_triplet_count1, '\0', 64 * sizeof(int));
  memset(gv_nb_triplet_count2, '\0', 64 * sizeof(int));
  memset(gv_nb_non_mutated_triplet_count, '\0', 64 * sizeof(int));
  memset(gv_nb_pair_of_triplets[0], '\0', 64 * 64 * sizeof(int));

  alignment_SG_stats_on_MA(dataquery, datachunktext, ma, 0,
			   gv_nb_transitions_transversions_indels, 
			   gv_nb_mutations_per_codon1, 
			   gv_nb_mutations_per_codon2);

  ma->entropy =  MIN3(entropyTriplet(gv_nb_non_mutated_triplet_count),
		      entropyTriplet(gv_nb_triplet_count1),
		      entropyTriplet(gv_nb_triplet_count2));

  ma->mutual  =  mutualInformationTriplet(gv_nb_pair_of_triplets);
  
  return ma->entropy < gp_entropy_min ;

}



/*
 * Entropy filter : remove low complexity regions
 */

int EntropyAndScoreFilter_MAList(Feature * f, char *dataquery, 
			 char *datatext, int *datatext_start)
{
    MA * ma = NULL, * ma_prev = NULL;

    ma = f->first_MA;

    /* (0) clean-up if nothing */
    if (!ma) {
      f->last_MA = NULL;
      return 0;
    }

    /* (1) find the first element OK and set it to *(p_firstMA) */
    while (EntropyAndScoreFilter_MA (f, ma, dataquery, datatext + datatext_start[ma->i_chunk])) {
      MA * to_free_ma = ma;
      ma = ma->next;
      FreeMA(to_free_ma);
      if (!ma) {
	f->first_MA = f->last_MA = NULL;
	return 0;
      }
    }    
    f->first_MA = f->last_MA = ma;    

    ma_prev = ma;
    ma = ma_prev->next;

    /* continue with the others MAs */
    while (ma) {
      if ( EntropyAndScoreFilter_MA(f, ma, dataquery, datatext + datatext_start[ma->i_chunk])) {
	  MA * to_free_ma = ma;
	  ma = ma->next;
	  ma_prev->next = ma;
	  FreeMA(to_free_ma);
	} else {
	  ma_prev = ma;
	  ma = ma->next;
	}
    }
    ma_prev->next = NULL;
    f->last_MA = ma_prev;
    
    return 0;
}






/*
 * Display for debuging effort
 */

int DisplayTuple(tuple * tuple)
{
  int left_correction = 0;
  printf("\t [(%9d,%9d),(%9d,%9d)] \t diag : %9d \t size : %9d \n", TBL_POS(tuple), TEL_POS(tuple), TBR_POS(tuple), TER_POS(tuple), tuple->diagonal , TSIZE(tuple));
  return 0;
}


int DisplayListTupleList(tuplelist * tuplelist)
{

  printf("**display tuple list**\n");
  while (tuplelist) {
    printf("\t--tl--\n");
    DisplayTupleList(tuplelist);
    tuplelist = tuplelist->next;
  }
  return 0;
}


int DisplayTupleList(tuplelist * tuplelist)
{
  tuple *tuple;
  if (tuplelist->first_tuple) {
    tuple = tuplelist->first_tuple;
    while (tuple) {
      DisplayTuple(tuple);
      tuple = tuple->next;
    }
    printf("\n");
  }
  return 0;
}

int DisplayGap(tuple * t1, tuple * t2)
{
  printf("\t intergaps {%d,%d}\n",TGAP_L(t1,t2),TGAP_R(t1,t2));
  assert(TGAP_L(t1,t2) >= 0);
  assert(TGAP_R(t1,t2) >= 0);
  /* if (TGAP_L(t1,t2)*TGAP_R(t1,t2) > MAX_WINDOW_REGROUP)
   *   printf("\t FIXME \n");
   */
  return 0;
}

int DisplayMA(MA * ma)
{
  int left_correction = 0;
  
  tuple * tuple = ma->first_tuple;
 
  while (tuple) {
    printf("\t");DisplayTuple(tuple);
    assert(TBL_POS(tuple)   >= ma->left_pos_begin);
    assert(TEL_POS(tuple)   <= ma->left_pos_end);
    assert(TBR_POS(tuple)   >= ma->right_pos_begin);
    assert(TBR_POS(tuple)   <= ma->right_pos_end);
    if (tuple->next){
      printf("\t");DisplayGap(tuple,tuple->next);
    }
    tuple = tuple->next;
  }
  printf("\t {(%d,%d), (%d,%d)}\n", ma->left_pos_begin, ma->left_pos_end, ma->right_pos_begin, ma->right_pos_end);
  printf("\n");
  return 0;
}

int DisplayListMA (MA * ma){

  while (ma){
    printf("\t MA : {(%d,%d),(%d,%d)}\n", ma->left_pos_begin, ma->left_pos_end, ma->right_pos_begin, ma->right_pos_end);
    
    /* ma attributes */
    assert(ma->reverse  <= 1);
    assert(ma->i_chunk  >= 0);
    assert(ma->i_chunk  < gp_nbchunks_text);
    assert (ma->blastscore >= 0);
    
    /* ma properties */
    assert(ma->left_pos_begin   >= 0);
    assert(ma->left_pos_begin   <= ma->left_pos_end);
    assert(ma->left_pos_end     <=  gp_querysize_chunk);
    assert(ma->right_pos_begin  >= 0);
    assert(ma->right_pos_begin  <= ma->right_pos_end);
    assert(ma->right_pos_end    <= gp_chunksize_text[ma->i_chunk]);
     
    /* ma tuples */ 
    DisplayMA(ma);
    
    ma = ma -> next;
  }

  return 0;
}

